/* 
  Copyright 2002 Andrew McNab mcnab@hep.man.ac.uk University of Manchester

  This file is part of GridSite.

  GridSite is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GridSite is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GridSite; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

/*------------------------------------------------------------------------*
 * For more information about GridSite: http://www.gridpp.ac.uk/gridsite/ *
 *------------------------------------------------------------------------*/

#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <gnome-xml/xmlmemory.h>
#include <gnome-xml/parser.h>

/* rh62? */
#include <sys/time.h>

#include "gridsite.h"
#include "libgridsite.h"

#define GRST_ETCPROXIES        "/../etc/proxies/"
#define GRST_SUBMITSPOOL       "/../etc/submitspool/"
#define GRST_SUBMITCONTACTS    "/../etc/submitcontacts/"
#define GRST_SUBMITERRORS      "/../etc/submiterrors/"
#define GRST_SUBMITSTATUS      "/../etc/submitstatus/"
#define GRST_PROXY_NO_PERM     "no_perm.txt"
#define GRST_PROXY_JOB_OK      "job_ok.txt"
#define GRST_PROXY_EXPIRED     "expired.txt"
#define GRST_PROXY_NOT_HTTPS   "not_https.txt"
#define GRST_PROXY_SPOOL_LIST  "spool_list.txt"
#define GRST_PROXY_OUTPUT_LIST "output_list.txt"
#define GRST_PROXY_STATUS      "job_status.txt"
#define GRST_PROXY_OUTPUT      "job_output.txt"

#define GLOBUSRUN      "/opt/globus/bin/globusrun"
#define GLOBUSRUNREDIR "/home/gridpp/gridsite/globusrun-redir"
#define GLOBUS_JOB_GET_OUTPUT "/opt/globus/bin/globus-job-get-output"

int ExecSystem(char *command, char *args[], char *environ[]) 
{
  int pid, status;

  if (command == NULL) return 1;

  pid = fork();
  if (pid == -1) return -1;
           
  if (pid == 0) 
    {
      dup2(1, 2);
      execve(command, args, environ);
      exit(127);
    }
           
  do {
       if (waitpid(pid, &status, 0) == -1) 
         {
           if (errno != EINTR) return -1;
         } 
       else return status;

     } while(1);
}

void ListJobs(char *certdn, char *submitspool)
{
  char         **array, *p, *hoststr, *jobmstr, 
                 *encodedDN, *decodedDN, *mildDN;
  int            n, i, my_n;
  DIR           *listDIR;
  struct dirent *oneentry;
  time_t         job_t;

  puts("<title>GridSite-Proxy: list waiting jobs</title>");    
  CatHeadOrFoot(GRST_HEADFILE);

  CatPathFile(requestdir, GRST_PROXY_SPOOL_LIST);
  
  listDIR = opendir(submitspool);

  n=0; 
  while ((oneentry = readdir(listDIR)) != NULL)
         if ((oneentry->d_name[0] != '.') && 
             (index(oneentry->d_name, '~') == NULL) &&
             (strcmp(oneentry->d_name, "core") != 0))  
           {
             /* printf("<!-- %d %s -->\n", n, oneentry->d_name); */
             ++n;
           }

  rewinddir(listDIR);

  array = (char **) malloc(sizeof(char *) * (n + 1));

  i=0;
  while (i <= n)
     {
       oneentry = readdir(listDIR);
       
       if (oneentry == NULL) break;
       
       if ((oneentry->d_name[0] != '.') && 
           (index(oneentry->d_name, '~') == NULL) &&
           (strcmp(oneentry->d_name, "core") != 0)) 
         {
           array[i] = strdup(oneentry->d_name);
           ++i;
         }
     } 
     
  array[i] = (char *) NULL;      

  closedir(listDIR);

  qsort((void *) array, (size_t) i, sizeof(char *), voidstrcmp);

  if (n == 0) puts("<p>No jobs awaiting submission.");
  else
    {
      my_n = 0;
      
      for (i=n-1; i >= 0; --i) 
         {
           p = index(array[i], ':');
           if (p == NULL) continue;
           
           p[0] = '\0';
           hoststr = &p[1];

           p = index(array[i], '.');
           if (p == NULL) continue;

           p[0] = '\0'; /* array[i] now starts with date as a string */

           job_t = 0;
           sscanf(array[i], "%ld", &job_t);

           p = index(hoststr, ':');
           if (p == NULL) continue;

           p[0] = '\0'; /* hoststr now is the target host name */

           jobmstr = &p[1];

           p = index(jobmstr, ':');
           if (p == NULL) continue;

           p[0] = '\0'; /* jobmstr now is the target jobmanager */

           encodedDN = &p[1];           

           decodedDN = UrlDecode(encodedDN);
           mildDN = MildUrlEncode(decodedDN);

           if (strcasecmp(certdn, decodedDN) == 0)
             {
               if (my_n == 0) puts("<dl>");

               printf("<dt><b>%s</b> %s\n<dd><small><a href=\"/website/members%s\">%s</a></small><p>\n", 
                   ctime(&job_t), hoststr, mildDN, decodedDN);
               ++my_n;
             }
                   
           free(decodedDN);
           free(mildDN);
           free(array[i]);
         }
                 
      if (my_n > 0) puts("</dl>");
      else          puts("<p>No jobs awaiting submission.");
    }
}

void ListOutputs(char *certdn, char *submitcontacts, char *submitstatus)
{
  char         **array, *p, *datestr, *hoststr, *encodedDN, *decodedDN, *mildDN,
                *statusfile;
  int            n, i, my_n, c;
  DIR           *listDIR;
  struct dirent *oneentry;
  time_t         job_t;
  FILE          *fp;

  puts("<title>GridSite-Proxy: list job outputs</title>");    
  CatHeadOrFoot(GRST_HEADFILE);

  CatPathFile(requestdir, GRST_PROXY_OUTPUT_LIST);
  
  listDIR = opendir(submitcontacts);

  n=0; 
  while ((oneentry = readdir(listDIR)) != NULL)
         if ((oneentry->d_name[0] != '.') && 
             (index(oneentry->d_name, '~') == NULL) &&
             (strcmp(oneentry->d_name, "core") != 0))  
           {
             /* printf("<!-- %d %s -->\n", n, oneentry->d_name); */
             ++n;
           }

  rewinddir(listDIR);

  array = (char **) malloc(sizeof(char *) * (n + 1));

  i=0;
  while (i <= n)
     {
       oneentry = readdir(listDIR);
       
       if (oneentry == NULL) break;
       
       if ((oneentry->d_name[0] != '.') && 
           (index(oneentry->d_name, '~') == NULL) &&
           (strcmp(oneentry->d_name, "core") != 0)) 
         {
           array[i] = strdup(oneentry->d_name);
           ++i;
         }
     } 
     
  array[i] = (char *) NULL;      

  closedir(listDIR);

  qsort((void *) array, (size_t) i, sizeof(char *), voidstrcmp);

  if (n == 0) puts("<p>No job outputs stored.");
  else
    {
      my_n = 0;

      for (i=n-1; i >= 0; --i) 
         {
           datestr = strdup(array[i]);

           p = index(datestr, ':');
           if (p == NULL) continue;
           
           p[0] = '\0';
           hoststr = &p[1];

           p = index(datestr, '.');
           if (p == NULL) continue;

           p[0] = '\0'; /* datestr now starts with date as a string */

           job_t = 0;
           sscanf(datestr, "%ld", &job_t);

           p = index(hoststr, ':');
           if (p == NULL) continue;

           p[0] = '\0'; /* hoststr now is the target host name */

           encodedDN = &p[1];           

           decodedDN = UrlDecode(encodedDN);
           mildDN = MildUrlEncode(decodedDN);

           if (strcasecmp(certdn, decodedDN) == 0)
             {
               if (my_n == 0) puts("<dl>");

               puts("<dt>");

               statusfile = malloc(strlen(submitstatus) + strlen(array[i]) + 2);
               sprintf(statusfile, "%s/%s", submitstatus, array[i]);
               
               fp = fopen(statusfile, "r");
               free(statusfile);
               
               if (fp != NULL)
                 {
                   while ((c = fgetc(fp)) != EOF) putchar(c);
                   fclose(fp);
                 }

               printf(" <b>%s</b> %s\n<dd><small>
                   <a href=\"/website/members%s\">%s</a></small>
                   <small>
                   <form action=gridsite-proxy.cgi method=post>
                   <input type=hidden name=job value=\"%s\">
                   <input type=submit name=action value=status>
                   <input type=submit name=action value=output>
                   </form></small>\n", 
                   ctime(&job_t), hoststr, mildDN, decodedDN, array[i]);
               ++my_n;
             }
                   
           free(decodedDN);
           free(mildDN);
           free(array[i]);
           free(datestr);
         }
                 
      if (my_n > 0) puts("</dl>");
      else          puts("<p>No job outputs stored.");
    }
}

void GetStatus(char *certdn, char *submitcontacts, char *submitstatus, 
               char *etcproxies)
{
  int   c;
  char *p, *jobstr, *contactfile, contact[512], *encodedDN,
       *args[5], *environ[5], *statusfile;
  FILE *fp;  

  puts("<title>GridSite-Proxy: show job status</title>");    
  CatHeadOrFoot(GRST_HEADFILE);

  jobstr = GetFromCGI("job");
  
  if (jobstr[0] == '\0') 
    {
      printf("Error: zero length jobstr\n");
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  p = index(jobstr, '/'); 
  if (p != NULL)
    {
      printf("Error: / in jobstr\n");
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  contactfile = malloc(strlen(submitcontacts) + strlen(jobstr) + 2);
  sprintf(contactfile, "%s/%s", submitcontacts, jobstr);
  
  fp = fopen(contactfile, "r");
  if (fp == NULL)  
    {
      printf("No such contact file %s (any more?)\n", contactfile);
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  CatPathFile(requestdir, GRST_PROXY_STATUS);

  if (fgets(contact, sizeof(contact), fp) == NULL)
    {
      printf("Submission failed (no contact string) - see Output\n");
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  fclose(fp);

  p = index(contact, '\n');
  if (p != NULL) *p = '\0';
  
  encodedDN = UrlEncode(certdn);

  printf("<!-- %s -->\n", contact);
  
  statusfile = malloc(strlen(submitstatus) + strlen(jobstr) + 2);
  sprintf(statusfile, "%s/%s", submitstatus, jobstr);

  fflush(stdout);

  args[0] = GLOBUSRUNREDIR;
  args[1] = statusfile;
  args[2] = "-status";
  args[3] = contact;
  args[4] = NULL;
  
  environ[0] = malloc(sizeof("X509_USER_PROXY=") + strlen(etcproxies) +
                      strlen(encodedDN) + 2);
  sprintf(environ[0], "X509_USER_PROXY=%s%s", etcproxies, encodedDN);

  environ[1] = "GLOBUS_LOCATION=/opt/globus";
  environ[2] = NULL;

  fflush(stdout);

  printf("<!-- %d -->\n", ExecSystem(GLOBUSRUNREDIR, args, environ));

  fp = fopen(statusfile, "r");
  if (fp == NULL)  
    {
      puts("Failed to read remote status");
    }
  else 
    {
      while ((c = fgetc(fp)) != EOF) putchar(c);
      fclose(fp);
    }

  CatHeadOrFoot(GRST_FOOTFILE);
}

void GetOutput(char *certdn, char *submitcontacts, char *submiterrors,
               char *etcproxies)
{
  char *p, *jobstr, *contactfile, contact[512], *encodedDN,
       *args[4], *environ[5];
  FILE *fp;  

  puts("<title>GridSite-Proxy: show job output</title>");    
  CatHeadOrFoot(GRST_HEADFILE);

  jobstr = GetFromCGI("job");
  
  if (jobstr[0] == '\0') 
    {
      printf("Error: zero length jobstr\n");
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  p = index(jobstr, '/'); 
  if (p != NULL)
    {
      printf("Error: / in jobstr\n");
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  CatPathFile(requestdir, GRST_PROXY_OUTPUT);

  puts("<p>Submission errors:\n<p>\n");
  if (!CatPathFile(GRST_SUBMITERRORS, jobstr)) puts("None");
  puts("\n<hr>\n");
  
  contactfile = malloc(strlen(submitcontacts) + strlen(jobstr) + 2);
  sprintf(contactfile, "%s/%s", submitcontacts, jobstr);
  
  fp = fopen(contactfile, "r");
  if (fp == NULL)  
    {
      printf("No such contact file %s\n", contactfile);
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  if (fgets(contact, sizeof(contact), fp) == NULL)
    {
      printf("No contact string\n");
      printf("<!-- %s -->\n", contactfile);
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  fclose(fp);

  p = index(contact, '\n');
  if (p != NULL) *p = '\0';
  
  encodedDN = UrlEncode(certdn);

  printf("<!-- %s -->\n", contact);
  
  fflush(stdout);

  environ[0] = malloc(sizeof("X509_USER_PROXY=") + strlen(etcproxies) +
                      strlen(encodedDN) + 2);
  sprintf(environ[0], "X509_USER_PROXY=%s%s", etcproxies, encodedDN);

  environ[1] = "GLOBUS_LOCATION=/opt/globus";
  environ[2] = NULL;

  args[0] = GLOBUS_JOB_GET_OUTPUT;
  args[1] = contact;
  args[2] = NULL;
  
  puts("Remote output:\n<p>\n");
  fflush(stdout);
  printf("<!-- %d -->\n", ExecSystem(GLOBUS_JOB_GET_OUTPUT, args, environ));
  puts("\n<hr>\n");

  args[0] = GLOBUS_JOB_GET_OUTPUT;
  args[1] = "-err";
  args[2] = contact;
  args[3] = NULL;
  
  puts("Remote errors:\n<p>\n");
  fflush(stdout);
  printf("<!-- %d -->\n", ExecSystem(GLOBUS_JOB_GET_OUTPUT, args, environ));
  puts("");

  CatHeadOrFoot(GRST_FOOTFILE);
}

void SubmitJob(char *certdn, char *submitspool)
{
  int     i;
  char    jobNNNN[8], *job, hostNNNN[9], *jobmanager, 
          jobmanagerNNNN[15], *host, *p,
          filename[512];
  time_t  this_t;
  struct  timeval tv;
  FILE   *fp;

  puts("<title>GridSite-Proxy: submiting job</title>");    
  CatHeadOrFoot(GRST_HEADFILE);

  time(&this_t);
  gettimeofday(&tv, (struct timezone *) NULL);

  for (i=0; i < 2000; ++i)
     {
       sprintf( jobNNNN,       "job%04d", i);
       sprintf(hostNNNN,       "host%04d", i);
       sprintf(jobmanagerNNNN, "jobmanager%04d", i);
 
       job  = GetFromCGI( jobNNNN);
       if (job[0] == '\0') continue;

       host       = GetFromCGI(hostNNNN);
       jobmanager = GetFromCGI(jobmanagerNNNN);
       if (jobmanager[0] == '\0') jobmanager = "jobmanager-fork";

/* buffer overrun here */
       sprintf(filename, "%s/%ld.%ld-%04d:%s:%s:%s", submitspool, 
          (long) this_t, (long) tv.tv_usec, i, host, jobmanager,
          UrlEncode(certdn));

       fp = fopen(filename, "w");
       if (fp == NULL) 
         {
           puts("Failed to open spool file\n");
           break;
         }
       
       for (p = job; *p != '\0'; ++p) if (*p != 0x0d) fputc(*p, fp);

       fclose(fp);
     }

  CatPathFile(requestdir, GRST_PROXY_JOB_OK);
}

main()
{
  char  *p, *q, *submitspool, *submitcontacts, *submiterrors, *submitstatus,
        *etcconf, *etcproxies, *certdn = NULL, *action;
  int        i, c, contentlen;
  xmlDocPtr  user_member;
  struct stat statbuf;
  FILE      *fp;

  puts("Content-type: text/html\n");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_NOTICE, "%s: connected", getenv("REMOTE_ADDR"));
#endif

  docroot = getenv("DOCUMENT_ROOT");
  requesturi = getenv("REQUEST_URI");
  if ((docroot == NULL) || (requesturi == NULL))
    {
      puts("Internal error"); 
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: empty docroot or requesturi", getenv("REMOTE_ADDR"));
      return;
    }

  pathtranslated = malloc(strlen(docroot) + strlen(requesturi) + 
                                        sizeof(GRST_INDEXHTML) + 1);
  strcpy(pathtranslated, docroot);
  strcat(pathtranslated, requesturi);
  
  if ((p = index(pathtranslated, '?'))  != NULL) *p = '\0';
  if ((p = index(pathtranslated, '#'))  != NULL) *p = '\0';
  if ((p = strstr(pathtranslated, "..")) != NULL) *p = '\0';

  requestdir = strdup(requesturi);
  p = rindex(requestdir, '/');
  if (p != NULL) p[1] = '\0';

  etcgroups  = malloc(strlen(docroot) + sizeof(GRST_ETCGROUPS) + 2);
  strcpy(etcgroups,  docroot);
  strcat(etcgroups,  GRST_ETCGROUPS);

  etcmembers = malloc(strlen(docroot) + sizeof(GRST_ETCMEMBERS) + 2);
  strcpy(etcmembers, docroot);
  strcat(etcmembers, GRST_ETCMEMBERS);

  etcproxies = malloc(strlen(docroot) + sizeof(GRST_ETCPROXIES) + 2);
  strcpy(etcproxies, docroot);
  strcat(etcproxies, GRST_ETCPROXIES);

  submitspool = malloc(strlen(docroot) + sizeof(GRST_SUBMITSPOOL) + 2);
  strcpy(submitspool, docroot);
  strcat(submitspool, GRST_SUBMITSPOOL);

  submitcontacts = malloc(strlen(docroot) + sizeof(GRST_SUBMITCONTACTS) + 2);
  strcpy(submitcontacts, docroot);
  strcat(submitcontacts, GRST_SUBMITCONTACTS);

  submitstatus = malloc(strlen(docroot) + sizeof(GRST_SUBMITSTATUS) + 2);
  strcpy(submitstatus, docroot);
  strcat(submitstatus, GRST_SUBMITSTATUS);

  submiterrors = malloc(strlen(docroot) + sizeof(GRST_SUBMITERRORS) + 2);
//  strcpy(submiterrors, docroot);
  strcpy(submiterrors, GRST_SUBMITERRORS);

  if (getenv("SSL_SERVER_CERT") != NULL) 
    { 
      ishttps = 1;
      certdn = getenv("SSL_CLIENT_S_DN"); /* trivial with certs! */
      if (certdn != NULL) userdn = certdn;
      ReadMemberData(&user_member, userdn);
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: SSL certdn is %s", getenv("REMOTE_ADDR"), certdn); 
    }
  else
    {
      puts("<title>GridSite-Proxy</title>");    
      CatHeadOrFoot(GRST_HEADFILE);
      CatPathFile(requestdir, GRST_PROXY_NOT_HTTPS);
      CatHeadOrFoot(GRST_FOOTFILE);
      return;
    }

  SetAclLevel();

  if (aclstatus == GRID_ACL_NONE)
    {
      puts("<title>GridSite-Proxy</title>");    
      CatHeadOrFoot(GRST_HEADFILE);
      CatPathFile(requestdir, GRST_PROXY_NO_PERM);
      CatHeadOrFoot(GRST_FOOTFILE);

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: permission denied", getenv("REMOTE_ADDR"));
#endif

      return;
    }

  contentlen = 0;
  p = getenv("CONTENT_LENGTH");
  if (p != NULL)
    { 
      sscanf(p, "%d", &contentlen);
      
      cgiposted = malloc(contentlen + 3);
      cgiposted[0] = '&';
      i = 1;
      while (i <= contentlen)
         {
           c = getchar();
           if (c == EOF) break;
           
           if (c != (int) 0x0d)
             {
               cgiposted[i] = c;
               ++i;
             }
         }

      cgiposted[i]   = '&';
      cgiposted[i+1] = '\0';
    }
  else cgiposted = "";
  
  action = GetFromCGI("action");
  
  if      (strcmp(action, "submit") == 0)  SubmitJob(certdn, submitspool);
  else if (strcmp(action, "outputs") == 0) ListOutputs(certdn, submitcontacts, submitstatus);
  else if (strcasecmp(action, "output") == 0)  
                  GetOutput(certdn, submitcontacts, submiterrors, etcproxies);
  else if (strcasecmp(action, "status") == 0)  
                  GetStatus(certdn, submitcontacts, submitstatus, etcproxies);
  else                       ListJobs(certdn, submitspool);

  CatHeadOrFoot(GRST_FOOTFILE);

  syslog(GRST_LOGLOCAL + LOG_NOTICE, 
              "%s: exit at end of main()", getenv("REMOTE_ADDR"));
}
