/* 
  Copyright 2002 Andrew McNab mcnab@hep.man.ac.uk University of Manchester

  This file is part of GridSite.

  GridSite is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GridSite is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GridSite; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

/*------------------------------------------------------------------------*
 * For more information about GridSite: http://www.gridpp.ac.uk/gridsite/ *
 *------------------------------------------------------------------------*/

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

#define GROUPSSUBDIR "groups"

void SplitCommaDN(char ***splitted, int *n, char *s)
{
  int   i;
  char *p, *q;

  p = strdup(s);
  
  *n = 1;
  for (q=p; *q != '\0'; ++q) if (*q == ',') ++(*n);

  *splitted = malloc(sizeof(char *) * (*n) + 1);

  i = 0;
  
  (*splitted)[i] = p;

  for (q=p; *q != '\0'; ++q)
   if (*q == ',') 
     {
       *q = '\0';
       ++i;
       (*splitted)[i] = &q[1];
     }

  (*splitted)[i+1] = NULL;
}

void SplitSlashDN(char ***splitted, int *n, char *s)
{
  int   i;
  char *p, *q;

  while ((*s != '/') && (*s != '\0')) ++s;
  
  p = strdup(s);
  
  *n = 0;
  for (q=p; *q != '\0'; ++q) if (*q == '/') ++(*n);

  *splitted = malloc(sizeof(char *) * (*n) + 1);

  i = 0;
  
  for (q=p; *q != '\0'; ++q)
   if (*q == '/') 
     {
       *q = '\0';
       (*splitted)[i] = &q[1];
       ++i;
     }

  (*splitted)[i] = NULL;
}

int main(int argn, char *argv[])
{
  int    i, ntopdn, nbasedn, nmemberdn;
  char   ldapline[513], **topdn, **basedn, *groupfile, groupline[513], *p,
       **memberdn, *base = NULL;
  FILE  *fp;

  if (argn < 3 )
    {
      puts("gridsite-vo TopDN GridSiteEtcDir");
      return 0;
    }
    
  SplitCommaDN(&topdn, &ntopdn, argv[1]);

  while (fgets(ldapline, sizeof(ldapline), stdin) != NULL)       
    if (strncasecmp(ldapline, "base: ", 6) == 0)
      {
        p = index(ldapline, '\n');
        if (p != NULL) *p = '\0';

        base = &ldapline[6];

        SplitCommaDN(&basedn, &nbasedn, base);
      
        if ((nbasedn == ntopdn + 1) && (strncasecmp(basedn[0], "ou=", 3) == 0))
          /* a group: try to list members */
          {
            p = index(&basedn[0][3], '/');
            if (p != NULL) *p = '\0'; /* avoids ../dir/file attacks */

            groupfile = malloc(strlen(argv[2]) + sizeof(GROUPSSUBDIR) +
                               strlen(&basedn[0][3]) + 3);
                               
            sprintf(groupfile,"%s/%s/%s",argv[2],GROUPSSUBDIR,&basedn[0][3]);


            fp = fopen(groupfile, "r");
            free(groupfile);
            if (fp != NULL)
              {
                printf("dn: %s\n", base);
                puts("objectClass: domain");
                puts("objectClass: organizationalUnit");
                puts("objectClass: groupofnames");

                while (fgets(groupline, sizeof(groupline), fp) != NULL)
                     {
                       p = index(groupline, '\n');
                       if (p != NULL) *p = '\0';
                       
                       SplitSlashDN(&memberdn, &nmemberdn, groupline);
                       
                       fputs("member: ", stdout);
                       
                       for (i=nmemberdn-1; i >= 0; --i)
                          {
                            fputs(memberdn[i], stdout);
                            if (i > 0) putchar(',');
                          }
                          
                       putchar('\n');
                     }
              
                fclose(fp);

                putchar('\n');

                puts("RESULT");
                puts("code: 0");
              }              
          }
        else
          /* assume a person: map comma DN to slash DN */
          {
            printf("dn: %s\n", &ldapline[6]);
            puts("objectClass: person");
            puts("objectClass: organizationalPerson");
            puts("objectClass: inetOrgPerson");
            puts("objectClass: pkiUser");
            fputs("description: subject= ", stdout);
            
            for (i=nbasedn-1; i >= 0; --i)
               {
                 putchar('/');
                 
                 for (p = basedn[i]; (*p != '\0') && (*p != '='); ++p)
                  putchar(*p);

                 while (*p != '\0')
                      { 
                        putchar(*p);
                        ++p;
                      }
                }
            
            puts("\n"); /* ie two \n */

            puts("RESULT");
            puts("code: 0");
          
          }
          
        free(basedn[0]);
        free(basedn);
      }
         
  return 0;
}
