/* 
  Copyright 2002 Andrew McNab mcnab@hep.man.ac.uk University of Manchester

  This file is part of GridSite.

  GridSite is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GridSite is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GridSite; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

/*------------------------------------------------------------------------*
 * For more information about GridSite: http://www.gridpp.ac.uk/gridsite/ *
 *------------------------------------------------------------------------*/

#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <gnome-xml/xmlmemory.h>
#include <gnome-xml/parser.h>

/* rh62? */
#include <sys/time.h>

#include "gridsite.h"
#include "libgridsite.h"

main()
{
  char  *p, *rawpage, *headtag, *startbody, *foottag, *q,
        *querystring, *mimetype, *cookiedn = NULL, *etcconf,
        *rawcookiedn = NULL, *passcode = NULL, *certdn = NULL,
         file_date[81];
  int        i, c, contentlen;
  xmlDocPtr  user_member;
  struct tm *thistm;
  struct stat statbuf;
  time_t     this_t;
  struct tm *file_tm;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: connected", getenv("REMOTE_ADDR"));
#endif

  docroot = getenv("DOCUMENT_ROOT");
  requesturi = getenv("REQUEST_URI");
  if ((docroot == NULL) || (requesturi == NULL))
    {
      puts("HTTP/1.1 500 Internal error");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\n\nInternal error"); 
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: empty docroot or requesturi", getenv("REMOTE_ADDR"));
      return;
    }

  pathtranslated = malloc(strlen(docroot) + strlen(requesturi) + 
                                        sizeof(GRST_INDEXHTML) + 1);
  strcpy(pathtranslated, docroot);
  strcat(pathtranslated, requesturi);
  
  if ((p = index(pathtranslated, '?'))  != NULL) *p = '\0';
  if ((p = index(pathtranslated, '#'))  != NULL) *p = '\0';
  if ((p = strstr(pathtranslated, "..")) != NULL) *p = '\0';

  if (pathtranslated[strlen(pathtranslated)-1] == '/') 
    {
      strcat(pathtranslated, GRST_INDEXHTML); /* try adding it on to / */
      if (stat(pathtranslated, &statbuf) != 0) 
        {
          p = rindex(pathtranslated, '/');
          p[1] = '\0'; /* strip GRST_INDEXHTML off again */
        }
    }       
  else if ((stat(pathtranslated, &statbuf) == 0) && S_ISDIR(statbuf.st_mode))
    {
        puts("HTTP/1.1 301 Moved");
        printf("Date: %s", asctime(thistm));
        printf("Location: %s/\n", requesturi);
        puts("Server: Apache (GridSite)\nContent-type: text/html\n\nYou need / on the end!");
        return;
    }

  etcgroups  = malloc(strlen(docroot) + sizeof(GRST_ETCGROUPS) + 2);
  strcpy(etcgroups,  docroot);
  strcat(etcgroups,  GRST_ETCGROUPS);

  etcmembers = malloc(strlen(docroot) + sizeof(GRST_ETCMEMBERS) + 2);
  strcpy(etcmembers, docroot);
  strcat(etcmembers, GRST_ETCMEMBERS);

  etcproxies = malloc(strlen(docroot) + sizeof(GRST_ETCPROXIES) + 2);
  strcpy(etcproxies, docroot);
  strcat(etcproxies, GRST_ETCPROXIES);

  etcconf    = malloc(strlen(docroot) + sizeof(GRST_ETCCONF) + 2);
  strcpy(etcconf,    docroot);
  strcat(etcconf,    GRST_ETCCONF);

  ReadEtcConf(etcconf);
  
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: REQUEST_URI %s; PATHTRANSLATED %s", 
                           getenv("REMOTE_ADDR"), requesturi, pathtranslated);
#endif

  requestdir = strdup(requesturi);
  p = rindex(requestdir, '/');
  if (p != NULL) p[1] = '\0';

  aclstatus = GRID_ACL_NONE;
  
  if ((StrCmpShort(requesturi, GRST_MEMBERS) == 0) ||
      (StrCmpShort(requesturi, GRST_GROUPS) == 0))  aclstatus = GRID_ACL_LIST;
  else if (strcmp(requesturi, GRST_STATUS) == 0)    aclstatus = GRID_ACL_READ;
  else if (stat(pathtranslated, &statbuf) != 0)
      {
        puts("HTTP/1.1 404 Not Found");
        printf("Date: %s", asctime(thistm));
        puts("Server: Apache (GridSite)\nContent-type: text/html\n\nPathtranslated not found");
#ifdef GRST_LOGLOCAL
        syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: absent file for request %s", 
              getenv("REMOTE_ADDR"), getenv("REQUEST_URI"));
#endif
        return;
      }

  if (getenv("SSL_SERVER_CERT") != NULL) 
    { 
      ishttps = 1;
      certdn = getenv("SSL_CLIENT_S_DN"); /* trivial with certs! */
      if (certdn != NULL) userdn = certdn;
      ReadMemberData(&user_member, userdn);
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: SSL certdn is %s", getenv("REMOTE_ADDR"), certdn);
 
      ReadCookies(&rawcookiedn, &passcode); /* try from cookies */
      if ((rawcookiedn != NULL) && (passcode != NULL))
        {
          ReadMemberData(&user_member, rawcookiedn);

          if (strcmp(passcode, GetElement(user_member, "passcode")) == 0) 
            {
              cookiedn = rawcookiedn; /* ie cookie+passcode correct */
              userdn   = rawcookiedn;
            }
        }

      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: after cookies userdn=%s", getenv("REMOTE_ADDR"), userdn);
    }

  if (strcmp(requesturi, GRST_STATUS) == 0)
    {
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: request for %s", getenv("REMOTE_ADDR"), GRST_STATUS);
      StatusPage(rawcookiedn, cookiedn, certdn, user_member);
      return;
    }

  if (aclstatus == GRID_ACL_NONE)
    {
      SetAclLevel();
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: SetAclLevel returns %d", getenv("REMOTE_ADDR"), aclstatus);
    }

  if (aclstatus == GRID_ACL_NOACL)
    {
      mimetype = GetMimeType(pathtranslated);
  
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: request has mimetype %s", getenv("REMOTE_ADDR"), 
               mimetype);
#endif

      puts("HTTP/1.1 200 OK");
      printf("Date: %s", asctime(thistm));

      stat(pathtranslated, &statbuf);
      file_tm = gmtime(&(statbuf.st_mtime));     
      strftime(file_date,sizeof(file_date),"%a, %d %b %Y %H:%M:%S %z",file_tm);				                                                                 
      printf("Last-Modified: %s\n", file_date);

      printf("Server: Apache (GridSite)\nContent-type: %s\n\n", mimetype);

      fp = fopen(pathtranslated, "r");
      if (fp != NULL) 
        {
          while ((c = fgetc(fp)) != EOF) putchar(c);
          fclose(fp);
        }

      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: no ACL so output directly", getenv("REMOTE_ADDR"));

      return;
    }

  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 No read permission");
      printf("Date: %s", asctime(thistm));

      CatHeadOrFoot(GRST_HEADFILE);
      printf("<table width=\"100%\" border=1 bgcolor=\"#FFFFFF\" cellpadding=5
       cellspacing=0><tr><td bgcolor=white><h1 align=center>Permission
       denied!</h1>The page you are trying to access is restricted.
       </td></tr></table>\n");

      puts("<p>");

      if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</td></tr></table></body>");

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: permission denied", getenv("REMOTE_ADDR"));
#endif

      return;
    }

  querystring = index(requesturi, '?');
  if (querystring != NULL) 
    {
      *querystring = '\0';    
      ++querystring;
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: querystring %s", getenv("REMOTE_ADDR"), querystring);
#endif
      if (strcmp(querystring, "upload") == 0) 
        {
          UploadFile();
          return;
        }
      else if (strcmp(querystring, "uploadproxy") == 0) 
        {
          UploadProxy();
          return;
        }
      else if (strcmp(querystring, "additem") == 0) 
        {          
          CreateItem();
          return;
        }
    }

  contentlen = 0;
  p = getenv("CONTENT_LENGTH");
  if (p != NULL)
    { 
      sscanf(p, "%d", &contentlen);
      
      cgiposted = malloc(contentlen + 3);
      cgiposted[0] = '&';
      for (i=1; i <= contentlen; ++i)
         {
           c = getchar();
           if (c == EOF) break;
           cgiposted[i] = c;                
         }

      cgiposted[i]   = '&';
      cgiposted[i+1] = '\0';
    }
  else cgiposted = "";
  
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
           "%s: cgiposted[]=%s", getenv("REMOTE_ADDR"), cgiposted);
#endif

  if (querystring != NULL) 
    {
      if      (strcmp(querystring, "sendpage") == 0) 
        {
          SendPage();
          return;
        }
      else if (strcmp(querystring, "senditem") == 0) 
        {
          SendItem();
          return;
        }
      else if (strcmp(querystring, "history") == 0) 
        {
          FileHistory();
          return;
        }
      else if (strcmp(querystring, "create") == 0) 
        {
          CreatePage();
          return;
        }
      else if (strcmp(querystring, "createdir") == 0) 
        {
          CreateDirectory();
          return;
        }
      else if (strcmp(querystring, "confirmdelete") == 0) 
        {
          ConfirmDelete();
          return;
        }
      else if (strcmp(querystring, "reallydelete") == 0) 
        {
          DeleteFile();
          return;
        }
      else if (strcmp(querystring, "list") == 0) 
        {
          ListDirectory(querystring);
          return;
        }
      else if (strcmp(querystring, "setcookie") == 0) 
        {
          SetCookie();
          return;
        }
      else if (strcmp(querystring, "editgroup") == 0) 
        {
          EditGroup();
          return;
        }
      else if (strcmp(querystring, "addgroup") == 0) 
        {
          AddGroup();
          return;
        }
      else if (strcmp(querystring, "editacl") == 0) 
        {
          EditAcl();
          return;
        }
      else if (strcmp(querystring, "addmember") == 0) 
        {
          AddMember();
          return;
        }
      else if (strcmp(querystring, "editmember") == 0) 
        {
          EditMember();
          return;
        }
    }
      
  if (StrCmpShort(requesturi, GRST_MEMBERS) == 0)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: request for %s", getenv("REMOTE_ADDR"), GRST_MEMBERS);
#endif
      p = &requesturi[strlen(GRST_MEMBERS)-1];

      if (p[1] == '\0')                     ListAllMembers();
      else                                  DisplayMember(p);

      return;
    }

  if (StrCmpShort(requesturi, GRST_GROUPS) == 0)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: request for %s", getenv("REMOTE_ADDR"), GRST_GROUPS);
#endif      
      p = &requesturi[strlen(GRST_GROUPS)];

      if (*p == '\0') ListAllGroups();
      else            ListGroup(p);

      return;
    } 

  mimetype = GetMimeType(pathtranslated);
  
  if (strcmp(mimetype, "text/x-weblog-item") == 0)
    {
      if ((querystring != NULL) && (strcmp(querystring, "edit") == 0))
           EditWeblogItem();
      else DisplayWeblogItem();
      return;
    }

  rawpage = ReadInPathTran();
  if (rawpage == NULL)
    {
      puts("HTTP/1.1 404 Not Found");
      printf("Date: %s", asctime(thistm));
      puts("Server: Apache (GridSite)\nContent-type: text/html\n\nCannot open pathtranslated");      
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: request %s is not found", getenv("REMOTE_ADDR"), 
               getenv("REQUEST_URI"));
      return;
    }

  if (querystring != NULL) 
    {
       if (strcmp(querystring, "edit") == 0) 
         {
           EditPage(rawpage, NULL, NULL);
           return;
         }

       if (strcmp(querystring, "print") == 0) 
         {
           DisplayPage(rawpage, 1);
           return;
         }

       if (*querystring != '\0') 
         {    
           puts("HTTP/1.1 404 Not Found");
           printf("Date: %s", asctime(thistm));
           printf("Server: Apache (GridSite)\nContent-type: text/html\n\nQuery %s not recognised",
             querystring);
           syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: query %s is not recognised", getenv("REMOTE_ADDR"), 
               querystring);
           return;
         }
    }
    
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: request has mimetype %s", getenv("REMOTE_ADDR"), 
               mimetype);
#endif

  if ((strcmp(mimetype, "text/html") == 0) && (*rawpage == '\0'))
       ListDirectory("");
  else if (pathtranslated[strlen(pathtranslated)-1] == '/')
       ListDirectory("");
  else if (strcmp(mimetype, "text/html") == 0)
       DisplayPage(rawpage, 0);
  else 
    {
      puts("HTTP/1.1 200 OK");
      printf("Date: %s", asctime(thistm));

      stat(pathtranslated, &statbuf);
      file_tm = gmtime(&(statbuf.st_mtime));     
      strftime(file_date,sizeof(file_date),"%a, %d %b %Y %H:%M:%S %z",file_tm);				                                                                 
      printf("Last-Modified: %s\n", file_date);

      printf("Server: Apache (GridSite)\nContent-type: %s\n\n", mimetype);

      fp = fopen(pathtranslated, "r");
      if (fp != NULL) 
        {
          while ((c = fgetc(fp)) != EOF) putchar(c);
          fclose(fp);
        }

     syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: non-HTML mimetype output directly", getenv("REMOTE_ADDR"));
    }
     
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: exit at end of main()", getenv("REMOTE_ADDR"));
}
