/* 
  Copyright 2002 Andrew McNab mcnab@hep.man.ac.uk University of Manchester

  This file is part of GridSite.

  GridSite is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  GridSite is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GridSite; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
*/

/*------------------------------------------------------------------------*
 * For more information about GridSite: http://www.gridpp.ac.uk/gridsite/ *
 *------------------------------------------------------------------------*/

#include <time.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <unistd.h>
#include <dirent.h>
#include <syslog.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <gnome-xml/xmlmemory.h>
#include <gnome-xml/parser.h>

/* rh62? */
#include <sys/time.h>

#include "gridsite.h"
int        ishttps = 0, aclstatus;
static     char nullstr[] = "";
char      *userdn     = nullstr,
          *docroot    = NULL,
          *etcgroups  = NULL,
          *etcmembers = NULL,
          *etcproxies = NULL,
          *pathtranslated = NULL, 
          *requestdir = NULL, 
          *requesturi = NULL, 
          *cgiposted;

xmlDocPtr  etcconfdoc;
xmlNodePtr member_elements = NULL;

/* yes, this will break every time we change the GRID_ACL_ symbols ...*/
static char *aclsyms[] = { "none",
                           "read",
                           "list",
                           "write",
                           "admin",
                           NULL     };


void ReadEtcConf(char *etcconf)
{
  xmlNodePtr  cur;
  
  etcconfdoc = xmlParseFile(etcconf);

  if (etcconfdoc == NULL) return;

  cur = xmlDocGetRootElement(etcconfdoc);
  
  if (xmlStrcmp(cur->name, (const xmlChar *) "gridsite-conf") != 0) return;

  cur = cur->xmlChildrenNode;

  if (xmlStrcmp(cur->name, (const xmlChar *) "member-elements") != 0) return;

  member_elements = cur->xmlChildrenNode;  
}

int xxvoidstrcmp(const char **p, const char **q)
/* used by directory qsort in SortDir() 
   Note the special order of sorting: down for digits, up for others */
{
  char *pp, *qq;
  
  pp = (char *) *p;
  qq = (char *) *q;

  while ((*pp == *qq) && (*pp != '\0')) /* find where they differ */
       {
         ++(*pp);
         ++(*qq);
       }
         
  if ((*pp == '\0') && (*qq == '\0')) return 0; /* identical */

  if (isdigit(*pp) && isdigit(*qq))
              if (*pp > *qq) return +1; /* reverse for digits */
              else           return -1;
  
  if (*pp < *qq) return -1; /* normal for others */
  else           return +1;
}

int voidstrcmp(const void *p, const void *q)
{
  char **pp, **qq;
  
  pp = (char **) p;
  qq = (char **) q; 
  
  return strcmp(*pp, *qq);
}


char **SortDir(char *dirpath)
{
  char         **array;
  int            n, i;
  DIR           *listDIR;
  struct dirent *oneentry;
  
  listDIR = opendir(dirpath);
  if (listDIR == NULL) return NULL;

  n=0; 
  while ((oneentry = readdir(listDIR)) != NULL)
         if ((oneentry->d_name[0] != '.') && 
             (index(oneentry->d_name, '~') == NULL) &&
             (strcmp(oneentry->d_name, "core") != 0))   ++n;

  rewinddir(listDIR);

  array = (char **) malloc(sizeof(char *) * (n + 1));

  i=0;
  while (i <= n)
     {
       oneentry = readdir(listDIR);
       
       if (oneentry == NULL) break;
       
       if ((oneentry->d_name[0] != '.') && 
           (index(oneentry->d_name, '~') == NULL) &&
           (strcmp(oneentry->d_name, "core") != 0)) 
         {
           array[i] = strdup(oneentry->d_name);
           ++i;
         }
     } 
     
  array[i] = (char *) NULL;      

  closedir(listDIR);

  qsort((void *) array, (size_t) i, sizeof(char *), voidstrcmp);

  return array;
}

int ConvertLevelSym(char *levelsym)
{
  int i;

  for (i=0; aclsyms[i] != NULL; ++i) 
   if (strcmp(levelsym, aclsyms[i]) == 0) return i;

  return 0;
}

char *GetFromCGI(char *name)
{
  char   *p, *namepattern, *valuestart, *returnvalue;
  int     c, i, j, n;  
  static  char *nullstring = "";
  size_t  size_needed;

  namepattern = malloc(strlen(name) + 3);
  sprintf(namepattern, "&%s=", name);
  
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
           "%s: looking for %s in POSTed data", 
           getenv("REMOTE_ADDR"), namepattern);
#endif

  p = strstr(cgiposted, namepattern);
  free(namepattern);
  if (p == NULL)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
           "%s: %s not present in POSTed data", getenv("REMOTE_ADDR"), name);
#endif
      return nullstring;
    }
     
  valuestart = &p[strlen(name) + 2];

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
           "%s: valuestart is %s in POSTed data", 
           getenv("REMOTE_ADDR"), valuestart);
#endif

  for (n=0; valuestart[n] != '&'; ++n) ;
  
  returnvalue = malloc(n + 1);
  
  j=0;
  
  for (i=0; i < n; ++i) 
     {
       if ((i < n - 2) && (valuestart[i] == '%')) /* url encoded as %HH */
         {
           returnvalue[j] = 0;
           
           if (isdigit(valuestart[i+1])) 
                 returnvalue[j] += 16 * (valuestart[i+1] - '0');
           else if (isalpha(valuestart[i+1])) 
                 returnvalue[j] += 16 * (10 + tolower(valuestart[i+1]) - 'a');
                         
           if (isdigit(valuestart[i+2])) 
                 returnvalue[j] += valuestart[i+2] - '0';
           else if (isalpha(valuestart[i+2])) 
                 returnvalue[j] += 10 + tolower(valuestart[i+2]) - 'a';
                                                   
           i = i + 2;
         }
       else if (valuestart[i] == '+') returnvalue[j] = ' ';
       else                           returnvalue[j] = valuestart[i];
       
       ++j;
     }

  returnvalue[j] = '\0';

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
       "%s: returning %s from GetFromCGI", getenv("REMOTE_ADDR"), returnvalue);
#endif

  return returnvalue;
}

char *ReadInPathTran(void)
{
  int     c;
  char   *rawpage, *p;
  size_t  rawsize;
  struct  stat buf;
  FILE   *fp;

  if (stat(pathtranslated, &buf) == -1) return NULL;

  rawsize = buf.st_size;
  rawpage = malloc(rawsize + (size_t) 1);
  if (rawpage == NULL) return NULL;

  fp = fopen(pathtranslated, "r");
  if (fp == NULL) return NULL;

  p = rawpage;
  while ((c = fgetc(fp)) != EOF)
       { 
         *p = c;
         ++p;
       }

  fclose(fp);
  
  *p = '\0';

  return rawpage;
}

FILE *DocRootfopen(char *filename, char *mode)
/* open file, where filename is relative to docroot (GridSite doc root) */
{
  char *fullfilename;
  FILE *fp;
  
  fullfilename = malloc(strlen(docroot) + 2 + strlen(filename));
  
  sprintf(fullfilename, "%s/%s", docroot, filename);
  
  fp = fopen(fullfilename, mode);
  
  free(fullfilename);
  
  return fp;
}

void CopyFile(char *destination, char *source)
{
  int    c;
  FILE *ifp, *ofp;

  ifp = fopen(source, "r");
  
  if (ifp != NULL)
    {
      ofp = fopen(destination, "w");
      if (ofp != NULL)
        {   
          while ((c = fgetc(ifp)) != EOF) fputc(c, ofp);    

          fclose(ofp);
        }
        
      fclose(ifp);
    }
}

int CatFile(char *filename)
{
  int    c;
  FILE *fp;

  fp = DocRootfopen(filename, "r");
  
  if (fp != NULL)
    {
      while ((c = fgetc(fp)) != EOF) putchar(c);    
      fclose(fp);
      return 1;
    }
  
  return 0;
}

int CatPathFile(char *pathname, char *filename)
{
  char  *pathfile; 
  int   c;
  FILE *fp;

  pathfile = malloc(strlen(pathname) + strlen(filename) + 2);
  sprintf(pathfile, "%s/%s", pathname, filename);

  fp = DocRootfopen(pathfile, "r");
  free(pathfile);
  
  if (fp != NULL)
    {
      while ((c = fgetc(fp)) != EOF) putchar(c);    
      fclose(fp);
      return 1;
    }
  
  return 0;
}

int CatHeadOrFoot(char *filename)
{
  int  found = 0;
  char *tmp_requestdir, *p;

  tmp_requestdir = strdup(requestdir);  

  while (!(found = CatPathFile(tmp_requestdir, filename)))
       {
         p = rindex(tmp_requestdir, '/');
         if (p == NULL) break;
         *p = '\0';
         p = rindex(tmp_requestdir, '/');
         if (p == NULL) break;
         p[1] = '\0';
       }
  
  free(tmp_requestdir);
  return found;
}

int UpdateFileLog(char *action, char *filename, int filesize, 
                                                 time_t now_t, long now_usec)
{
  int   c;
  char *logfilename, *tmplogfilename;
  FILE *oldfp, *newfp;

  logfilename    = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 6);
  tmplogfilename = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 31);
  sprintf(logfilename, "%s%s.%s.log", docroot, requestdir, filename);
  sprintf(tmplogfilename, "%s%s.%s.%ld.%06ld.log", docroot, requestdir, 
                                 filename, (long) now_t, (long) now_usec);

#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
       "%s: logfilename=%s tmplogfilename=%s", getenv("REMOTE_ADDR"), 
       logfilename, tmplogfilename);
#endif

  newfp = fopen(tmplogfilename, "w");
  if (newfp == NULL) return 0;

  oldfp = fopen(logfilename, "r");
  if (oldfp != NULL)
    {
      while ((c = fgetc(oldfp)) != EOF) fputc(c, newfp);
      fclose(oldfp);
    }

  fprintf(newfp, "%s %ld %ld %d %s\n", 
                 action, filesize, now_t, now_usec, userdn);
  fclose(newfp);

  rename(tmplogfilename, logfilename);
  return 1;
}

void GetFileDate(char **httpdate, char **htmldate)
{
  int    c;
  char  *p, *logfilename, *requestdir, *filename, oneline[161], lastline[161],
         s[81];
  static char httpdate_s[81], htmldate_s[81];
  FILE  *fp;
  time_t file_t;
  struct tm *file_tm;

  httpdate_s[0] = '\0';
  htmldate_s[0] = '\0';

  *httpdate = httpdate_s;
  *htmldate = htmldate_s;

  requestdir = strdup(requesturi);
  
  filename = rindex(requestdir, '/');
  *filename = '\0';
  ++filename;
  if (*filename == '\0') filename = "index.html";

  logfilename = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 8);

  sprintf(logfilename, "%s/%s/.%s.log", docroot, requestdir, filename);

  fp = fopen(logfilename, "r");
  if (fp != NULL)
    {
      while (fgets(oneline, sizeof(oneline), fp) != NULL)
           {
             p = rindex(oneline, '\n');
             if (p != NULL) *p = '\0';
             strcpy(lastline, oneline);
           }
     
      fclose(fp);

      sscanf(lastline, "%*s %*d %ld", &file_t);
      file_tm = localtime(&file_t);     
      strftime(htmldate_s, sizeof(htmldate_s), "%a %e %B %Y", file_tm); 
      strftime(httpdate_s, sizeof(httpdate_s), "%a, %d %b %Y %H:%M:%S %z", 
                                                                 file_tm); 
    }

  free(requestdir);
}

char *GetMimeType(char *filename)
/* return mime type associated with the . extension of filename, or 
   GRST_DEFAULTMIME if none can be found. associations are searched for in
   GRST_ETCMIME (which will normally be /etc/mime.types)              */
{
  int          i;
  char        *p, *ext, *extspaces;
  static char  type[513];
  FILE        *fp;

  ext = rindex(filename, '.');
  if (ext == NULL) 
    {
      strcpy(type, GRST_DEFAULTMIME);
      return type;
    }

  ++ext;
  
  fp = fopen(GRST_ETCMIME, "r");
  if (fp != NULL)
    {
      extspaces = malloc(strlen(ext) + 3);
      sprintf(extspaces, " %s ", ext);

      while (fgets(type, sizeof(type) - 1, fp) != NULL)
        {
          p = index(type, '#');
          if (p != NULL) *p = '\0';
          
          strcat(type, " ");
          for (i=0; type[i] != '\0'; ++i) if (isspace(type[i])) type[i] = ' ';

          if (strstr(type, extspaces) != NULL)
            {
              fclose(fp);
              free(extspaces);

              for (i=0; !isspace(type[i]); ++i) ;
              type[i] = '\0';
          
              return type;
            }
        }

      free(extspaces);
      fclose(fp);
    }
    
  strcpy(type, GRST_DEFAULTMIME);
  return type;
}

char *UrlDecode(char *in)
{
  int   i, j, n;
  char *out;
  
  n = strlen(in);
  out = malloc(n + 1);
  
  j=0;
  
  for (i=0; i < n; ++i) 
     {
       if ((i < n - 2) && (in[i] == '%')) /* url encoded as %HH */
         {
           out[j] = 0;
           
           if (isdigit(in[i+1])) 
                 out[j] += 16 * (in[i+1] - '0');
           else if (isalpha(in[i+1])) 
                 out[j] += 16 * (10 + tolower(in[i+1]) - 'a');
                         
           if (isdigit(in[i+2])) 
                 out[j] += in[i+2] - '0';
           else if (isalpha(in[i+2])) 
                 out[j] += 10 + tolower(in[i+2]) - 'a';
                                                   
           i = i + 2;
         }
       else if (in[i] == '+') out[j] = ' ';
       else                   out[j] = in[i];
       
       ++j;
     }

  out[j] = '\0';
  
  return out;
}

char *UrlEncode(char *in)
{
  char *out, *p, *q;
  
  out = malloc(3*strlen(in) + 1);
  
  p = in;
  q = out;
  
  while (*p != '\0')
       {
         if (isalnum(*p) || (*p == '.') || (*p == '=') || (*p == '-'))
           {
             *q = *p;
             ++q;
           }
         else
           {
             sprintf(q, "%%%2x", *p);
             q = &q[3];
           }

         ++p;
       }
  
  *q = '\0';  
  return out;
}

char *MildUrlEncode(char *in)
{
  char *out, *p, *q;
  
  out = malloc(3*strlen(in) + 1);
  
  p = in;
  q = out;
  
  while (*p != '\0')
       {
         if (isalnum(*p) || (*p == '.') || (*p == '=') || (*p == '-') 
                         || (*p == '/') || (*p == '@') || (*p == '_'))
           {
             *q = *p;
             ++q;
           }
         else
           {
             sprintf(q, "%%%2x", *p);
             q = &q[3];
           }

         ++p;
       }
  
  *q = '\0';  
  return out;
}

void HtmlPutc(char c)
{
  if      (c == '&') fputs("&amp;" , stdout);
  else if (c == '<') fputs("&lt;"  , stdout);
  else if (c == '>') fputs("&gt;"  , stdout);
  else if (c == '"') fputs("&quot;", stdout);
  else               putchar(c);  
}

void HtmlPuts(char *s)
/* output string s to stdout, escaping HTML codes and converting HTTP URL's to
   links. "URL's" begin http:// or www and end with whitespace. */
{
  int n = 0, m, len, anychars = 1;
  
  len = strlen(s);
  
  while (s[n] != '\0')
   {
     if (!isspace(s[n])) anychars = 1;
     else if (s[n] == '\n')
       {
         if (!anychars) puts("<p>");
         anychars = 0;
       }

     if ((tolower(s[n]) != 'w') && (tolower(s[n]) != 'h')) 
       {
         HtmlPutc(s[n]);
         ++n;
         continue;
       }
       
     if (strncasecmp(&s[n], "http://", (size_t) 7) == 0)
       {
         fputs("<a href=\"", stdout);
         for (m = n; (!isspace(s[m]) && (s[m] != '\0')); ++m) putchar(s[m]);
         fputs("\">", stdout);
         for (m = n; (!isspace(s[m]) && (s[m] != '\0')); ++m) putchar(s[m]);
         fputs("</a>", stdout);          
         n = m;
         continue;
       }

     if (strncasecmp(&s[n], "www.", (size_t) 4) == 0)
       {
         fputs("<a href=\"http://", stdout);
         for (m = n; (!isspace(s[m]) && (s[m] != '\0')); ++m) putchar(s[m]);
         fputs("\">", stdout);
         for (m = n; (!isspace(s[m]) && (s[m] != '\0')); ++m) putchar(s[m]);
         fputs("</a>", stdout);          
         n = m;
         continue;
       }
     
     putchar(s[n]);
     ++n;
   }    
}

int ReadMemberData(xmlDocPtr *thisone, char *memberdn)
{
  char *personpath, *p, *memberdnurl;

  if (memberdn == NULL) return 0;
  
  memberdnurl = UrlEncode(memberdn);
  personpath  = malloc(strlen(etcmembers) + strlen(memberdnurl) + 1);
  
  strcpy(personpath, etcmembers);
  strcat(personpath, memberdnurl);
 
  *thisone = xmlParseFile(personpath);

  if (*thisone == NULL) return 0;

  return 1;
}

char *GetElement(xmlDocPtr member, char *element)
{
  xmlNodePtr  cur;
  xmlChar    *content;
  
  cur = xmlDocGetRootElement(member);

// do we really need this check? if we leave it out, this routine is general  
//  if (xmlStrcmp(cur->name, (const xmlChar *) "member") != 0) return nullstr;

  cur = cur->xmlChildrenNode;
  
  while (cur != NULL)
       {
         if (xmlStrcmp(cur->name, (const xmlChar *) element) == 0)
           {
             content = xmlNodeGetContent(cur);

             if (content == NULL) return nullstr;
             else                 return content;
           }
           
         cur = cur->next;
       }                                           

  return nullstr;
}

void PutElement(xmlDocPtr member, char *element, char *content)
{
  xmlNodePtr root, cur;
  
  root = xmlDocGetRootElement(member);

  cur = root->xmlChildrenNode;
  
  while (cur != NULL)
       {
         if (xmlStrcmp(cur->name, (const xmlChar *) element) == 0)
           {
             xmlNodeSetContent(cur, content);
             return;
           }
           
         cur = cur->next;
       }
       
  xmlNewTextChild(root, NULL, element, content);
}

int InGroup(char *group, char *dn)
/* returns GRID_ACL_READ if dn[] is in group group[] and
   GRID_ACL_NONE if not (in future, will return GRID_ACL_WRITE etc if
   dn[] should be allowed to modify the group list etc) */
{
  int   i, level = GRID_ACL_NONE;
  char *p, *groupfile, dnline[GRST_ROLELEN+GRST_DNLEN+2], *groupdn;
  FILE *groupfp;

  for (p = group; *p != '\0'; ++p)
    if (!isalnum(*p) && (*p != '@') && (*p != '.') && (*p != ':')) 
      return GRID_ACL_NONE;
  
  groupfile = malloc(strlen(etcgroups) + 
                     strlen(group) + 1);
         
  strcpy(groupfile, etcgroups);
  strcat(groupfile, group); 

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: process groupfile %s in InGroup()", 
               getenv("REMOTE_ADDR"), groupfile);
#endif

  groupfp = fopen(groupfile, "r");
  free(groupfile);

  if (groupfp != NULL)
    {
      while (fgets(dnline, GRST_ROLELEN+GRST_DNLEN+2, groupfp) != NULL)
           {
             p = index(dnline, '\n');
             if (p != NULL) *p = '\0';

             p = index(dnline, ' ');
             if (p != NULL)
               {
                 *p = '\0';
                 groupdn = &p[1];

#ifdef GRST_LOGLOCAL
   syslog(GRST_LOGLOCAL + LOG_DEBUG, 
               "%s: in groupfile in InGroup() %s %s", 
               getenv("REMOTE_ADDR"), dnline, groupdn);
#endif
                 if (strcmp(groupdn, dn) == 0)
                   {
                     for (i=0; aclsyms[i] != NULL; ++i)
                      if (strcmp(aclsyms[i], dnline) == 0) 
                        {
                          level = i;
                          break;
                        }
                   }
               }
           }
             
      fclose(groupfp);           
    }         

  return level;
}

int InAcl(char *group)
/* return acl level of group in the ACL of the current directory */
{
  char *aclfile, *p, oneline[GRST_ROLELEN + GRST_GROUPLEN + 2], *aclgroup;
  int   i;
  FILE *aclfp;

  aclfile = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 1);
  strcpy(aclfile, pathtranslated);
  
  p = rindex(aclfile, '/');
  if (p == NULL) return GRID_ACL_NONE; /* weirdness */  
  p[1] = '\0';
  
  strcat(aclfile, GRST_ACL_FILE);
 
  aclfp = fopen(aclfile, "r");
  free(aclfile);
  if (aclfp == NULL) return GRID_ACL_NONE;

  while (fgets(oneline, GRST_ROLELEN + GRST_GROUPLEN + 2, aclfp) != NULL)
       {
         p = index(oneline, '\n');
         if (p != NULL) *p = '\0';

         p = index(oneline, ' ');
         if (p != NULL)  
           { 
             *p = '\0';
             aclgroup = &p[1];
           }
         else continue;
         
         if (strcmp(aclgroup, group) == 0)
           for (i=0; aclsyms[i] != NULL; ++i)
              if (strcmp(oneline, aclsyms[i]) == 0)
                {
                  fclose(aclfp);
                  return i;
                } 
       }
    
  fclose(aclfp);
  return GRID_ACL_NONE;
}

void SetAclLevel(void)
/* returns GRID_ACL_WRITE, GRID_ACL_READ or GRID_ACL_NONE
   about *userdn based on the ACL in the directory of this page  */
{
  char *aclfile, *p, oneline[GRST_ROLELEN + GRST_GROUPLEN + 2], *dn,
       *aclgroup;
  int   level;
  FILE *aclfp;
  
  aclstatus = GRID_ACL_NONE;

  aclfile = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 1);
  
  strcpy(aclfile, pathtranslated);
  
  p = rindex(aclfile, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(aclfile, GRST_ACL_FILE);
      
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: Try to open %s in SetAclLevel()", 
               getenv("REMOTE_ADDR"), aclfile);
#endif

  aclfp = fopen(aclfile, "r");
  free(aclfile);
  if (aclfp != NULL)
    {
      if (InGroup(GRST_EDITDIRS, userdn)) aclstatus = GRID_ACL_ADMIN;
      else
        {
          aclstatus = GRID_ACL_NONE;

          while (fgets(oneline,GRST_ROLELEN+GRST_GROUPLEN+2,aclfp) != NULL)
            {
              p = index(oneline, '\n');
              if (p != NULL) *p = '\0';

              p = index(oneline, ' ');
              if (p != NULL)  
                { 
                  *p = '\0';
                  aclgroup = &p[1];
                }
              else continue;

              if (strcmp(aclgroup, GRID_ACL_ANY_DN) == 0)
                {
                  level = ConvertLevelSym(oneline);
                  if (level > aclstatus) aclstatus = level;
                }
              else if (InGroup(aclgroup, userdn) > GRID_ACL_NONE) 
                {
                  level = ConvertLevelSym(oneline);
                  if (level > aclstatus) aclstatus = level;
                }
            }
        }
        
      fclose(aclfp);
    }
  else 
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: Cannot open a %s in SetAclLevel() %s", 
               getenv("REMOTE_ADDR"), GRST_ACL_FILE, strerror(errno));
#endif
      aclstatus = GRID_ACL_NOACL;
    }
    
  return;
}

void ReadCookies(char **cookiedn, char **passcode)
{
  int   i;
  char *p, *q, *cookies_userdnurl, *cookies_usercode;

  *cookiedn = NULL;
  *passcode = NULL;
  
  p = getenv("HTTP_COOKIE");
  if (p == NULL) return;

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: HTTP cookies offered: %s", getenv("REMOTE_ADDR"), p);
#endif

  cookies_userdnurl = strdup(p);
  cookies_usercode  = strdup(p);

  p = strstr(cookies_userdnurl, "gridsiteDN=");
  if (p != NULL)
    {
      *cookiedn = malloc(strlen(cookies_userdnurl)+1);
      p = &p[11]; /* 11 = strlen("gridsiteDN=") */
      q = *cookiedn;
      while ((*p != ';') && (*p != '\0'))
           {
             if (p[0] == '%') if (isalnum(p[1])) if (isalnum(p[2]))
               {
                 sscanf(&p[1], "%2x", &i);
                 *q = i;
                 ++q;
                 p = &p[3];
                 continue;
               }

             *q = *p;
             ++q;
             ++p;
           }
           
      *q = '\0';
    }
  
  p = strstr(cookies_usercode, "gridsiteCODE=");
  if (p != NULL)
    {
      *passcode = &p[13]; /* 13 = strlen("gridsiteCODE=") */
      while ((*p != ';') && (*p != '\0')) ++p;
      *p = '\0';
    }
}

void NewsHeadLines(void)
/* this is a quick hack to try this idea out: if the user puts <news> 
   in their HTML document, we follow it with the latest weblog item's
   date and title from the directory /news/ !!! */
{
  int        last, i;
  char      *p, *dirpath, **dirarray, *itempath, *datestr, htmldate[81];
  time_t     item_t = 0;
  struct tm *item_tm;
  xmlDocPtr  thisitem;
  
  dirpath = malloc(strlen(docroot) + 7);
  strcpy(dirpath, docroot);
  strcat(dirpath, "/news/");
  dirarray = SortDir(dirpath);

  if (dirarray == NULL) return;

  for (last=0; dirarray[last] != NULL; ++last) ;

  for (i=last-1; i >= 0; --i)
    {
      itempath = malloc(strlen(dirpath) + 1 + strlen(dirarray[i]));
      strcpy(itempath, dirpath);
      strcat(itempath, dirarray[i]);

      datestr = strdup(dirarray[i]);
      p = index(datestr, '.');
      if (p == NULL) 
        {
          free(datestr);
          free(itempath);
          continue;
        }
  
      *p = '\0';
  
      sscanf(datestr, "%ld", &item_t);
      free(datestr);
  
      if (item_t == 0) 
        {
          free(itempath);
          continue;
        }

      printf("<!-- .%s -->\n", rindex(itempath, '/'));
      thisitem = xmlParseFile(itempath);
      free(itempath);

      if (thisitem == NULL) continue; /* failed to open it ok */

      item_tm = localtime(&item_t);
      strftime(htmldate, sizeof(htmldate), "%a %e %B %Y", item_tm);

      printf("%s:\n%s\n", htmldate, GetElement(thisitem, "title"));

/*    puts("<br>");
      HtmlPuts(GetElement(thisitem, "intro")); */

      free(thisitem);
      free(dirpath);
      return;
    }
    
  free(dirpath);
}

void DisplayPage(char *rawpage, int forprinting)
{
  char   *p, *headtag, *startbody, *foottag, *q, *httpdate, *htmldate;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Permission denied.");
      printf("Date: %s\nPermission denied.\n", asctime(thistm));
      return;
    }

  GetFileDate(&httpdate, &htmldate);  

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  if (httpdate[0] != '\0') printf("Last-Modified: %s\n", httpdate);

  puts("Server: GridSite\nContent-type: text/html\n");

  headtag = strstr(rawpage, "<body");
  if (headtag == NULL) 
    { 
      headtag = strstr(rawpage, "<BODY");
      if (headtag == NULL) headtag = strstr(rawpage, "<Body");
    }

  if (headtag != NULL)
    {
      for (p = rawpage; p < headtag; ++p) putchar(*p);
      while ((*p != '>') && (*p != '\0')) ++p;
      if (*p == '>') ++p;
      startbody = p;
    }
  else startbody = rawpage;

  if (forprinting) 
    {
      if (!CatHeadOrFoot(GRST_HEADFILE_PRINT)) puts("<body>");
    }
  else 
    {
      if (!CatHeadOrFoot(GRST_HEADFILE)) puts("<body>");
    }

  foottag = strstr(startbody, "</body>");
  if (foottag == NULL) 
    { 
      foottag = strstr(startbody, "</BODY>");
      if (foottag == NULL) 
        {
          foottag = strstr(startbody, "</Body>");
          if (foottag == NULL) foottag = &startbody[strlen(startbody)];
        }
    }

  puts("<!-- start of user contributed HTML -->");
    
  for (p = startbody; p < foottag; ++p) 
     {
       /* start this as a hack, and do something more general in the future */
       if (*p == '<') if (strncasecmp(p, "<news>", 6) == 0)
         {
           NewsHeadLines();
         }

       putchar(*p);
     }

  puts("<!-- end of user contributed HTML -->");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: page body sent", getenv("REMOTE_ADDR"));
#endif

  puts("<p><small>");

  if (htmldate[0] != '\0')
    {
      printf("<hr>Last updated %s\n . ", htmldate);
      
      if (!forprinting) 
        printf("<a href=\"%s?history\">View page history</a>", requesturi);
    }

  puts("<hr>");
  
  if (userdn[0] != '\0')
    {
      printf("You are %s<br>\n", userdn); 

      if ((aclstatus >= GRID_ACL_WRITE) && (!forprinting))
          if (requesturi != NULL)
           printf("<a href=\"%s?edit\">Edit&nbsp;page</a> .\n", requesturi);

      if ((aclstatus >= GRID_ACL_LIST) && (!forprinting))
          puts("<a href=\"./?list\">List&nbsp;directory</a> .");
    }   
 
  if (!forprinting)
    {
      if (ishttps) 
           printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
      else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

      printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

      printf("<a href=\"%s?print\">Print&nbsp;View</a> .\n", requesturi);
    }
  else printf("<a href=\"%s\">Normal&nbsp;View</a> .\n", requesturi);
     
  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n",
             GRST_VERSION);
     
  puts("</small>");

  free(rawpage);

  if (forprinting)
    {
      if (!CatHeadOrFoot(GRST_FOOTFILE_PRINT)) puts("</body>");
    }
  else
    {
      if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</body>");
    }

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

int BareWeblogItem(char *itempath, int includebody)
{
  char   *p, *datestr, htmldate[81], *author, *mildauthor,
         *forenames, *surname;
  time_t     item_t = 0;
  struct tm *item_tm;
  xmlDocPtr  thisitem, thismember;
      
  p = rindex(itempath, '/');
  if (p == NULL) return 0;
  datestr = strdup(&p[1]);
  p = index(datestr, '.');
  if (p == NULL) 
    {
      free(datestr);
      return 0;
    }
  
  *p = '\0';
  
  sscanf(datestr, "%ld", &item_t);
  free(datestr);
  
  if (item_t == 0) return 0;

  thisitem = xmlParseFile(itempath);

  if (thisitem == NULL) return 0; /* failed to open it ok */

  printf("<!-- .%s -->\n", rindex(itempath, '/'));

  printf("<b><big>%s</big></b>\n", GetElement(thisitem, "title"));
  
  item_tm = localtime(&item_t);
  strftime(htmldate, sizeof(htmldate), "%a %e %B %Y %H:%M %Z", item_tm);

  printf("<br><small>%s.\n", htmldate);

  fputs("Posted&nbsp;by&nbsp;", stdout);

  author = GetElement(thisitem, "author");
  if (!ReadMemberData(&thismember, author))
    {
      puts(author);
    }
  else
    {
      mildauthor = MildUrlEncode(author);
      printf("<a href=\"%s%s\">", GRST_MEMBERS, &mildauthor[1]);
      free(mildauthor);

      forenames = GetElement(thismember, "forenames");
      surname   = GetElement(thismember, "surname");
      
      if ((forenames[0] == '\0') &&
          (surname[0] == '\0')) 
        {
          for (p = author; *p != '\0'; ++p)
           if (isspace(*p)) fputs("&nbsp;", stdout);
           else putchar(*p);
        }
      else 
        {
          for (p = forenames; *p != '\0'; ++p)
           if (isspace(*p)) fputs("&nbsp;", stdout);
           else HtmlPutc(*p);

          fputs("&nbsp;", stdout);
      
          for (p = surname; *p != '\0'; ++p)
           if (isspace(*p)) fputs("&nbsp;", stdout);
           else HtmlPutc(*p);
        }
              
      puts("</a>");
    }

  puts("</small><br>");

  if (includebody) puts("<p>");

  HtmlPuts(GetElement(thisitem, "intro"));

  if (includebody) 
    {
      puts("<p>");
      HtmlPuts(GetElement(thisitem, "body"));    
    }
  else if (*(GetElement(thisitem, "body")) != '\0') 
     printf("<br><b><a href=\".%s\"><small>Read more...</small></a></b>\n", rindex(itempath, '/'));
  else 
     printf("<br><b><a href=\".%s\"><small>View item</small></a></b>\n", rindex(itempath, '/'));
  
  free(thisitem);

  return 1; /* success */
}

void DisplayWeblogItem(void)
{
  char   *p, *headtag, *startbody, *foottag, *q, *httpdate, *htmldate;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Permission denied.");
      printf("Date: %s\nPermission denied.\n", asctime(thistm));
      return;
    }
 
  GetFileDate(&httpdate, &htmldate);

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  if (httpdate[0] != '\0') printf("Last-Modified: %s\n", httpdate);

  puts("Server: GridSite\nContent-type: text/html\n");

  printf("<title>Weblog %s</title>\n", htmldate);

  if (!CatHeadOrFoot(GRST_HEADFILE)) puts("<body>");

  puts("<p><a href=\"./\"><small>Back to the weblog</small></a><p>");

  if (!BareWeblogItem(pathtranslated, 1))
    {
      puts("Failed to open weblog item.\n");
      return;      
    }
//  else puts("<p align=right><a href=\"./\"><small>Back to the weblog</small></a>");

  puts("<p><small>");

  if (htmldate[0] != '\0')
      printf("<hr>Last updated %s\n . 
        <a href=\"%s?history\">View item history</a>", htmldate, requesturi);

  puts("<hr>");
  
  if (userdn[0] != '\0')
    {
      printf("You are %s<br>\n", userdn); 

      if (aclstatus >= GRID_ACL_WRITE)
          if (requesturi != NULL)
           printf("<a href=\"%s?edit\">Edit&nbsp;item</a> .\n", requesturi);

      if (aclstatus >= GRID_ACL_LIST)
          puts("<a href=\"./?list\">List&nbsp;directory</a> .");
    }   

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n",
             GRST_VERSION);
     
  puts("</small>");

  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</body>");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: item footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void EditWeblogItem(void)
{
  int        i;
  char      *p, *headtag, *startbody, *foottag, *q, *filename,
            *datestr, htmldate[81], *author, *mildauthor,
            *forenames, *surname;
  time_t     item_t = 0;
  struct tm *item_tm, *thistm;
  time_t     this_t;
  xmlDocPtr  thisitem, thismember;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Permission denied.");
      printf("Date: %s\nPermission denied.\n", asctime(thistm));
      return;
    }
 
  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: GridSite\nContent-type: text/html\n");

  puts("<title>Edit weblog item</title>");

  if (!CatHeadOrFoot(GRST_HEADFILE)) puts("<body>");

  p = rindex(pathtranslated, '/');
  if (p == NULL) return;
  filename = &p[1];
  datestr = strdup(filename);
  p = index(datestr, '.');
  if (p == NULL) 
    {
      puts("bad format date");
      free(datestr);
      return;
    }
  
  *p = '\0';
  
  sscanf(datestr, "%ld", &item_t);
  free(datestr);
  
  if (item_t == 0) 
    {
      puts("item has zero time!");
      return;
    }

  thisitem = xmlParseFile(pathtranslated);

  if (thisitem == NULL) 
    {
      puts("failed to open/parse!");
      return; /* failed to open it ok */
    }

  printf("<form action=\"%s?senditem\" method=post>\n", requestdir);  

  printf("<input type=hidden name=filename value=\"%s\">\n", filename);

  puts("<table>");

  puts("<tr><td colspan=2><input type=submit value=\"Save changes\"></td></tr>");

  item_tm = localtime(&item_t);
  strftime(htmldate, sizeof(htmldate), "%a %e %B %Y %H:%M %Z", item_tm);

  printf("<tr><td>Date:</td><td>%s</td>\n", htmldate);

  puts("<tr><td>Posted by:</td><td>");

  author = GetElement(thisitem, "author");
  if (strcmp(userdn, author) == 0) puts(author);
  else
    { 
      puts("<select name=author>");
      printf("<option selected>%s\n", author);
      printf("<option>%s\n", userdn);
      puts("</select>");
    }
    
  puts("</td></tr>");
  
  printf("<tr><td>Title:</td>
          <td><input type=text name=title value=\"%s\" size=40></td></tr>\n", 
                                      GetElement(thisitem, "title"));
 
  fputs("<tr><td>Introduction:</td>
        <td><textarea name=intro cols=40 rows=3>", stdout);        
  p = GetElement(thisitem, "intro"); 
  i = strlen(p);
  if (p[i] == '\n') p[i] = '\0';
  while (*p != '\0') 
       {
         HtmlPutc(*p);    
         ++p;
       }
  puts("</textarea></td></tr>\n"); 

  puts("</table>");

  fputs("<p><textarea name=body rows=10 cols=60>", stdout);
  p = GetElement(thisitem, "body"); 
  i = strlen(p);
  if (p[i] == '\n') p[i] = '\0';
  while (*p != '\0') 
       {
         HtmlPutc(*p);
         ++p;
       }
  puts("</textarea>\n"); 
                
  free(thisitem);

  puts("<tr><td colspan=2><input type=submit value=\"Save changes\"></td></tr>");

  puts("</form>");

  if (!CatHeadOrFoot(GRST_FOOTFILE)) puts("</body>");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: edit item footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void ListAllMembers()
{
  int        i;
  char      *persondn, **dirarray;
  struct tm *thistm;
  time_t     this_t;
  xmlDocPtr  thisperson;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: GridSite\nContent-type: text/html\n");

  puts("<title>List all website members</title>");
    
  CatHeadOrFoot(GRST_HEADFILE);

  puts("<h1>List all website members</h1>");

  puts("<table>");
  dirarray = SortDir(etcmembers);
  if (dirarray != NULL)
    {
      for (i=0; dirarray[i] != NULL; ++i)
         {
           persondn = UrlDecode(dirarray[i]);

           if (ReadMemberData(&thisperson, persondn))
             {
                printf("<tr><td>%s %s</td>
                      <td><small><a href=\"%s\">%s</a></small></td></tr>\n", 
                        GetElement(thisperson, "forenames"), 
                        GetElement(thisperson, "surname"), 
                        MildUrlEncode(&persondn[1]),
                        persondn);
                free(thisperson);
             }
           else printf("<tr><td>&nbsp;</td><td><small>%s</small></td></tr>\n",
                        persondn);

           free(persondn);
         }
    }
    
  puts("</table>");

  if ((InGroup(GRST_ADDMBRS,  userdn) > GRID_ACL_NONE) ||
      (InGroup(GRST_EDITMBRS, userdn) > GRID_ACL_NONE))
    printf("<p><form action=\"%s?addmember\" method=post>
            <small><input name=dn size=60>
            <input type=submit value=\"Add\"></small></form>\n",
            GRST_MEMBERS);
          
  puts("<p><small>");
  puts("<hr>");
  
  if (userdn[0] != '\0') printf("You are %s<br>\n", userdn); 

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");

  CatHeadOrFoot(GRST_FOOTFILE);

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void DisplayMember(char *rawpersondn)
{
  int         ngroups = 0, i, editgrps, c, anycmdout;
  char       *persondn, **dirarray, *p, *forenames, *surname, 
             *encodeduserdn, *proxyfile, *cmd;
  struct tm  *thistm;
  time_t      this_t;
  struct stat buf;
  FILE       *fp;
  xmlDocPtr   thisperson;
  xmlNodePtr  cur;
  xmlChar    *content, *text, *type, *maxlength;
  
  time(&this_t);
  thistm = gmtime(&this_t); 
      
  if (aclstatus == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Permission denied.");
      printf("Date: %s\nPermission denied.\n", asctime(thistm));
      return;
    }

  persondn = UrlDecode(rawpersondn);

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: GridSite\nContent-type: text/html\n");

  printf("<title>%s</title>\n", persondn);

  CatHeadOrFoot(GRST_HEADFILE);

  if (ReadMemberData(&thisperson, persondn))
    {
      puts("<p><table>");
      
      printf("<tr><td>DN:</td><td>%s</td></tr>\n", persondn);

      forenames = GetElement(thisperson, "forenames");
      surname   = GetElement(thisperson, "surname");
      
      if ((forenames[0] != '\0') || (surname[0] != '\0'))
           printf("<tr><td>Name:</td><td>%s %s</td></tr>\n",forenames,surname);

      for (cur = member_elements; cur != NULL; cur = cur->next)
           {
             content = xmlNodeGetContent(cur);
             if (content == NULL) continue;
             if (xmlStrcmp(content,(const xmlChar *) "forenames")==0) continue;
             if (xmlStrcmp(content,(const xmlChar *) "surname") == 0) continue;

             text    = xmlGetProp(cur, (const xmlChar *) "text");
             type    = xmlGetProp(cur, (const xmlChar *) "type");

             p = GetElement(thisperson, content);
             if (p[0] != '\0')
               {
                 if (text == NULL) printf("<tr><td>%s:</td>\n", content);
                 else              printf("<tr><td>%s:</td>\n", text);

                 if (type == NULL)
                  printf("<td>%s</td></tr>\n", p);
                 else if (xmlStrcmp(type, (const xmlChar *) "email") == 0)
                  printf("<td><a href=\"mailto:%s\">%s</a></td></tr>\n", p, p);
                 else if (xmlStrcmp(type, (const xmlChar *) "url") == 0)
                  printf("<td><a href=\"%s\">%s</a></td></tr>\n", p, p);
                 else
                  printf("<td>%s</td></tr>\n", p);
               }
           }

      puts("<tr><td>Groups:</td><td>");
      
      ngroups = 0;
      dirarray = SortDir(etcgroups);
      if (dirarray != NULL)
        {
          for (i=0; dirarray[i] != NULL; ++i)
             if (InGroup(dirarray[i], persondn) > GRID_ACL_NONE)
               {
                 if (ngroups > 0) fputs(", ", stdout);

                 printf("<a href=\"%s%s\">%s</a>", 
                              GRST_GROUPS, dirarray[i], dirarray[i]);
                 ++ngroups;
               }

          free(dirarray);
        }

      if (ngroups == 0) puts("None");
      puts("</td></tr>");

      puts("</table>\n");      
    }
  else printf("<p>No such user: %s\n", persondn);

  /* Let viewer edit this member if they are the member or are in the
     right groups */

  if ((strcmp(userdn, persondn) == 0) ||
      (InGroup(GRST_EDITMBRS, userdn) > GRID_ACL_NONE) ||
      ((InGroup(GRST_ADDMBRS, userdn) > GRID_ACL_NONE) &&
       (strcmp(userdn, GetElement(thisperson, "addedby")) == 0)) )
    {
      printf("<form action=\"%s?editmember\" method=post>\n<table>\n", 
              GRST_MEMBERS);

      printf("<input type=hidden name=dn value=\"%s\">\n", persondn);

      printf("<tr><td>Forenames:</td>
          <td><input name=forenames value=\"%s\" size=30 maxlength=%d></td></tr>\n",
              GetElement(thisperson, "forenames"), GRST_DEFMAXLEN);

      printf("<tr><td>Surname:</td>
          <td><input name=surname value=\"%s\" size=30 maxlength=%d></td></tr>\n",
              GetElement(thisperson, "surname"), GRST_DEFMAXLEN);

      for (cur = member_elements; cur != NULL; cur = cur->next)
           {
             content = xmlNodeGetContent(cur);
             if (content == NULL) continue;
             if (xmlStrcmp(content,(const xmlChar *) "forenames")==0) continue;
             if (xmlStrcmp(content,(const xmlChar *) "surname") == 0) continue;

             text      = xmlGetProp(cur, (const xmlChar *) "text");
             type      = xmlGetProp(cur, (const xmlChar *) "type");
             maxlength = xmlGetProp(cur, (const xmlChar *) "maxlength");

             if (text == NULL) printf("<tr><td>%s:</td>\n", content);
             else              printf("<tr><td>%s:</td>\n", text);

             if (maxlength == NULL)
                   printf("<td><input name=\"%s\" value=\"%s\" size=30 
                           maxlength=%d></td></tr>\n",
                           content,
                           GetElement(thisperson, content), GRST_DEFMAXLEN);
             else
                   printf("<td><input name=\"%s\" value=\"%s\" size=30 
                           maxlength=\"%s\"></td></tr>\n",
                           content,
                           GetElement(thisperson, content), maxlength);               
           }

      puts("<tr><td align=center colspan=2><input type=submit 
                 value=\"Update details\"></td></tr>\n</form>");

      if ((strcmp(userdn, persondn) == 0) && 
          (stat(etcproxies, &buf) == 0))
        {
          puts("<tr><td colspan=2>&nbsp;</td></tr>");

          if (stat(GRST_GRIDPROXYINFO, &buf) == 0)
            {
              encodeduserdn = UrlEncode(userdn);
              proxyfile = malloc(strlen(etcproxies) + strlen(encodeduserdn) + 1);
              strcpy(proxyfile, etcproxies);
              strcat(proxyfile, encodeduserdn);
                                      
              if (stat(proxyfile, &buf) == 0)
                {
                  cmd = malloc(sizeof(GRST_GRIDPROXYINFO) + strlen(proxyfile)
                               + 12);
                  sprintf(cmd, "%s -file %s -all", GRST_GRIDPROXYINFO,
                               proxyfile);
                  fp = popen(cmd, "r");
                  if (fp != NULL)
                    {
                      anycmdout = 0;
                    
                      while ((c = getc(fp)) != EOF)
                           {
                             if (!anycmdout)
                               {
                                 puts("<tr><td>Current Proxy:</td><td><small>");
                                 anycmdout = 1;
                               }
                                                 
                             if (c == '\n') puts("; ");
                             else putchar(c);
                           }
                      
                      pclose(fp);
                      if (anycmdout) puts("</small></td></tr>\n");
                    }
                }
            }

          printf("<form action=\"%s?uploadproxy\" method=post
                   enctype=\"multipart/form-data\">\n", GRST_MEMBERS);
          puts("<tr><td>Upload Proxy:</td>");
          puts("<td><input type=file name=localfile></td></tr>");
          puts("<tr><td align=center colspan=2><input type=submit 
                                   value=\"Upload Proxy\"></td></tr></form>"); 
        }

      puts("</table>");
    }
  
  /* Let viewer add this person to groups they are admin of */

  editgrps = InGroup(GRST_EDITGRPS, userdn);
  ngroups = 0;      
  dirarray = SortDir(etcgroups);
  if (dirarray != NULL)
    {
      for (i=0; dirarray[i] != NULL; ++i)
         if (InGroup(dirarray[i], persondn) == GRID_ACL_NONE)
            if ((InGroup(dirarray[i], userdn) >= GRID_ACL_WRITE) ||
                (editgrps > GRID_ACL_NONE))
              {
                if (ngroups == 0) 
                 printf("<center><form action=\"%s?editgroup\" method=post><hr>
                         Add member to groups:
                         <select name=group>\n", GRST_GROUPS);

                printf("<option>%s\n", dirarray[i]);

                ++ngroups;
              }

      free(dirarray);
    }

  if (ngroups > 0) 
       printf("</select>
               <input type=hidden name=levelname value=read>
               <input type=hidden name=dn value=\"%s\">
               <input type=submit value=\"Add\"></form></center>",
               persondn);

  puts("<hr><small>");
  
  if (userdn[0] != '\0')
    {
      printf("You are %s<br>\n", userdn); 

      if (aclstatus >= GRID_ACL_WRITE)
          if (requesturi != NULL)
           printf("<a href=\"%s?edit\">Edit&nbsp;details</a> .\n", requesturi);
    }   

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);
       
  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n",GRST_MEMBERS);
  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");

  CatHeadOrFoot(GRST_FOOTFILE);

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void ListGroup(char *groupname)
{
  char      *persondn, *p, *groupfile, line[513], *groupnamedup, *dn,
            *infofile;
  struct tm *thistm;
  time_t     this_t;
  FILE      *fp;
  xmlDocPtr  thisperson;

  time(&this_t);
  thistm = gmtime(&this_t); 

  if (InGroup(GRST_EDITGRPS, userdn) > GRID_ACL_NONE) aclstatus=GRID_ACL_ADMIN;
  else aclstatus = InGroup(groupname, userdn);
          
  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: GridSite\nContent-type: text/html\n");

  groupnamedup = strdup(groupname);
  for (p=groupnamedup; (*p != '\0') && isalnum(*p); ++p) ;
  *p = '\0'; /* force end of string, possibly early */

  printf("<title>Group: %s</title>", groupnamedup);
    
  CatHeadOrFoot(GRST_HEADFILE);

  infofile = malloc(sizeof(GRST_GROUPSINFO) + strlen(groupnamedup) + 6);
  sprintf(infofile, "%s/%s.txt", GRST_GROUPSINFO, groupnamedup);

  if (!CatFile(infofile)) printf("<h1>Group: %s</h1>", groupnamedup);
  free(infofile);

  groupfile = malloc(strlen(etcgroups) + 2 + strlen(groupnamedup));
  sprintf(groupfile, "%s/%s", etcgroups, groupnamedup);

  free(groupnamedup);
  
  fp = fopen(groupfile, "r");
  if (fp != NULL)
    {
      puts("<p><table><tr><th>Name</th><th>User DN</th>
                          <th>Access</th></tr>");

      while (fgets(line, sizeof(line), fp) != NULL)
           {
             dn = index(line, ' ');
             if (dn != NULL)
               {
                 *dn = '\0';
                 ++dn;                 
               }
             else dn = "error";
             
             p = index(dn, '\n');
             if (p != NULL) *p = '\0';

             if (aclstatus < GRID_ACL_ADMIN) puts("<tr>"); 
             else printf("<form action=\"%s?editgroup\" method=post><tr>",
                                                              GRST_GROUPS);

             if (ReadMemberData(&thisperson, dn))             
                  printf("<td>%s %s</td>
                          <td><small><a href=\"%s%s\">%s</a></small></td>\n", 
                          GetElement(thisperson, "forenames"), 
                          GetElement(thisperson, "surname"), 
                          GRST_MEMBERS, 
                          MildUrlEncode(&dn[1]), 
                          dn);
             else printf("<td>&nbsp;</td><td><small>%s</small></td>\n", dn);

             if (aclstatus < GRID_ACL_ADMIN)
               {
                 if (strcmp(line, aclsyms[GRID_ACL_ADMIN]) == 0)
                      printf("<td align=center>%s</td><td>&nbsp;</td>\n", 
                                   aclsyms[GRID_ACL_ADMIN]);
                 else puts("<td align=center>member</td><td>&nbsp;</td>");
               }
             else
               {
                 puts("<td align=middle><small><select name=levelname>");

                 if (strcmp(line, aclsyms[GRID_ACL_ADMIN]) == 0)
                      printf("<option value=\"%s\" selected>admin
                              <option value=\"%s\">member
                              <option value=\"%s\">remove\n", 
                              aclsyms[GRID_ACL_ADMIN], aclsyms[GRID_ACL_READ],
                                                       aclsyms[GRID_ACL_NONE]);
                 
                 else printf("<option value=\"%s\">admin
                              <option value=\"%s\" selected>member
                              <option value=\"%s\">remove\n", 
                              aclsyms[GRID_ACL_ADMIN], aclsyms[GRID_ACL_READ],
                                                       aclsyms[GRID_ACL_NONE]);
                 
                 printf("<input type=hidden name=group value=\"%s\">
                         <input type=hidden name=dn value=\"%s\">
                         </select>\n", groupname, dn);

                 puts("<input type=submit value=Change>");
                 puts("</small></td>");
               }
             
             if (aclstatus < GRID_ACL_ADMIN) puts("</tr>"); 
             else                            puts("</tr></form>");
           }
       
      fclose(fp);

      puts("</table>\n");
    }
  
  puts("<p><small>");

  puts("<hr>");
  
  if (userdn[0] != '\0') printf("You are %s<br>\n", userdn); 

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);
  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");

  CatHeadOrFoot(GRST_FOOTFILE);

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void ListAllGroups(void)
{
  int        i;
  char      *persondn, *dirpath, **dirarray;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 
      
  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));

  puts("Server: GridSite\nContent-type: text/html\n");

  puts("<title>List all groups</title>");
    
  CatHeadOrFoot(GRST_HEADFILE);

  puts("<h1>List all groups</h1>");

  dirarray = SortDir(etcgroups);

  if (dirarray != NULL)
    {
      for (i=0; dirarray[i] != NULL; ++i)
        printf("<p align=center><a href=\"./%s\">%s</a>\n", 
                                       dirarray[i], dirarray[i]);
      free(dirarray);
    }

  if (InGroup(GRST_EDITGRPS, userdn) > GRID_ACL_NONE)
    printf("<center><form action=\"%s?addgroup\" method=post>
            <small><input name=group size=%d>
            <input type=submit value=\"Add group\"></small></form></center>\n",
            GRST_MEMBERS, GRST_GROUPLEN);
          
  puts("<p><small>");

  puts("<hr>");
  
  if (userdn[0] != '\0') printf("You are %s<br>\n", userdn); 

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);

  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");

  CatHeadOrFoot(GRST_FOOTFILE);

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: page footer sent", getenv("REMOTE_ADDR"));
#endif

  return;
}

void EditGroup(void)
{
  int        fd, i;
  char      *p, *levelname, *group, *fullgrouptemp, *fullgroup, oneline[513],
            *groupdn, *dn, *aclaccess;
  struct tm *thistm;
  time_t     this_t;
  FILE      *groupfp, *tmpfp;
    
  time(&this_t);
  thistm = gmtime(&this_t); 

  dn = GetFromCGI("dn");

  group = GetFromCGI("group");

  if (InGroup(GRST_EDITGRPS, userdn) > GRID_ACL_NONE) aclstatus=GRID_ACL_ADMIN;
  else aclstatus = InGroup(group, userdn);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have write permission to that group");
      return;
    }

  levelname = GetFromCGI("levelname");
  for (i=0; aclsyms[i] != NULL; ++i)
    if (strcmp(aclsyms[i], levelname) == 0) break;

  if (aclsyms[i] == NULL)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      printf("Level %s doesnt exist", levelname);
      return;
    }
    
  fullgrouptemp = malloc(strlen(etcgroups) + strlen(group) + 9);
  sprintf(fullgrouptemp, "%s/%s.XXXXXX", etcgroups, group);

  fd = mkstemp(fullgrouptemp);
  if (fd == -1) return;
  close(fd);
  
  tmpfp = fopen(fullgrouptemp, "w");
  if (tmpfp == NULL) 
    {
      free(fullgrouptemp);
      return;
    }

  if (strcmp(levelname, "none") != 0) fprintf(tmpfp, "%s %s\n", levelname, dn);

  fullgroup = malloc(strlen(etcgroups) + strlen(group) + 2);
  sprintf(fullgroup, "%s/%s", etcgroups, group);

  groupfp = fopen(fullgroup, "r");
  if (groupfp == NULL) 
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      printf("Group %s doesnt exist\n", group);
      return;
    }

  while (fgets(oneline, sizeof(oneline), groupfp) != NULL)
           {
             /* now attempt to parse this line: ACCESSLEVEL "DN" */

             aclaccess = oneline;

             while (isspace(*aclaccess)) ++aclaccess;  /* skip leading spaces */
             p = aclaccess;

             while (!isspace(*p) && (*p != '\0')) ++p; /* find end of level */

             if (*p == '\0') continue; /* must be a space: skip line */
         
             *p = '\0';
         
             ++p;
          
             while (isspace(*p)) ++p;
         
             groupdn = p;

             p = index(groupdn, '\n');   /* get rid of tedious trailing \n */
             if (p != NULL) *p = '\0';

             if (*groupdn == '\0') continue; /* DN must exist */

             if (strcasecmp(groupdn, dn) != 0)
                      fprintf(tmpfp, "%s %s\n", aclaccess, groupdn); 
           }

  fclose(groupfp);
  fclose(tmpfp);

  rename(fullgrouptemp, fullgroup);

  free(fullgroup);
  free(fullgrouptemp);
    
  /* everything worked ok so we just redirect to the directory list */
  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite");

  printf("Location: %s%s\n\n", GRST_GROUPS, group);
}

void EditAcl(void)
{
  int        fd, i;
  char      *p, *levelname, *group, *fullacltemp, *fullacl, oneline[513],
            *aclgroup, *aclaccess;
  struct tm *thistm;
  time_t     this_t;
  FILE      *aclfp, *tmpfp;
    
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_ADMIN)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have admin permission for that ACL");
      return;
    }

  group = GetFromCGI("group");

  levelname = GetFromCGI("levelname");
  for (i=0; aclsyms[i] != NULL; ++i)
    if (strcmp(aclsyms[i], levelname) == 0) break;

  if (aclsyms[i] == NULL)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      printf("Level %s doesnt exist", levelname);
      return;
    }
    
  fullacltemp = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 13);
  
  strcpy(fullacltemp, pathtranslated);
  
  p = rindex(fullacltemp, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(fullacltemp, ".tmp");
  strcat(fullacltemp, GRST_ACL_FILE);
  strcat(fullacltemp, ".XXXXXX");

  fd = mkstemp(fullacltemp);
  if (fd == -1) return;
  close(fd);
  
  tmpfp = fopen(fullacltemp, "w");
  if (tmpfp == NULL) 
    {
      free(fullacltemp);
      return;
    }

  if (strcmp(levelname, "none") != 0) 
       fprintf(tmpfp, "%s %s\n", levelname, group);


  fullacl = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 2);
  strcpy(fullacl, pathtranslated);
  
  p = rindex(fullacl, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(fullacl, GRST_ACL_FILE);

  aclfp = fopen(fullacl, "r");
  if (aclfp != NULL) 
    {    
      while (fgets(oneline, sizeof(oneline), aclfp) != NULL)
           {
             /* now attempt to parse this line: ACCESSLEVEL GROUP */

             aclaccess = oneline;

             while (isspace(*aclaccess)) ++aclaccess;  /* skip leading spaces */
             p = aclaccess;

             while (!isspace(*p) && (*p != '\0')) ++p; /* find end of level */

             if (*p == '\0') continue; /* must be a space: skip line */
         
             *p = '\0';
         
             ++p;
          
             while (isspace(*p)) ++p;
         
             aclgroup = p;

             p = index(aclgroup, '\n'); /* get rid of tedious trailing \n */
             if (p != NULL) *p = '\0';

             if (*aclgroup == '\0') continue; /* group must be given */

             if (strcasecmp(aclgroup, group) != 0)
                      fprintf(tmpfp, "%s %s\n", aclaccess, aclgroup); 
           }
      fclose(aclfp);
    }

  fclose(tmpfp);

  rename(fullacltemp, fullacl);

  free(fullacl);
  free(fullacltemp);
    
  /* everything worked ok so we just redirect to the directory list */
  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite");

  printf("Location: %s?list#groups\n\n", requestdir);
}

void ConfirmDelete(void)
{
  int        n;
  char      *p;
  struct tm *thistm;
  time_t     this_t;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have write permission to that directory");
      return;
    }

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>Confirm delete of");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<h1>Confirm delete of");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  puts("<p><font size=+5>Do you really want to delete this file?");
         
  fputs("<p align=center><a href=\"", stdout);
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("?reallydelete\">YES!</a>");

  requestdir = strdup(requesturi);
  p = rindex(requestdir, '/');
  p[1] = '\0';
  
  printf("<p align=center><a href=\"%s?list\">NO!</a>\n", requestdir);

  CatHeadOrFoot(GRST_FOOTFILE);
}

void DeleteFile(void)
{
  int        n, savederrno;
  char      *p, *filename;
  struct tm *thistm;
  time_t     this_t;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have write permission to that directory");
      return;
    }
    
  if (remove(pathtranslated) == 0)
    {
      filename = rindex(pathtranslated, '/');
      ++filename;
      UpdateFileLog("Delete", filename, 0, this_t, 0);

      /* everything worked ok so we just redirect to the directory list */
      puts("HTTP/1.1 302 Moved Temporarily");
      printf("Date: %s", asctime(thistm));
      fputs("Server: GridSite\nLocation: ", stdout);

      fputs(requestdir, stdout);
      puts("?list");
      return;      
    }

  savederrno = errno;

  puts("HTTP/1.1 500 Denied");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>Cannot delete ");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Cannot delete ");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  printf("<p>rmdir() failed with error %d: %s\n", 
            savederrno, strerror(savederrno));
         
  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
}

void AddGroup(void)
{
  char        *p, *group, *grouppath;
  struct tm   *thistm;
  time_t       this_t;
  struct stat  statbuf;
  FILE        *fp;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (InGroup(GRST_EDITGRPS, userdn) == GRID_ACL_NONE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have permission to add groups");
      return;
    }
    
  group = GetFromCGI("group");
// check for bad characters
  
  grouppath = malloc(strlen(etcgroups) + strlen(group) + 2);
  sprintf(grouppath, "%s/%s", etcgroups, group);

  if (stat(grouppath, &statbuf) != 0)
    {
      fp = fopen(grouppath, "w");
      if (fp == NULL)
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: GridSite\nContent-type: text/html\n");      

          puts("Failed to create record for new group");
          return;
        }
      else fclose(fp);
    }

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite");
  printf("Location: %s\n\n", GRST_MEMBERS);
}

void AddMember(void)
{
  char        *p, *dn, *dnencoded, *memberpath;
  struct tm   *thistm;
  time_t       this_t;
  struct stat  statbuf;
  FILE        *fp;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if ((InGroup(GRST_ADDMBRS,  userdn) == GRID_ACL_NONE) &&
      (InGroup(GRST_EDITMBRS, userdn) == GRID_ACL_NONE))
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have permission to add members");
      return;
    }
    
  dn = GetFromCGI("dn");
  dnencoded = UrlEncode(dn);  
  
  memberpath = malloc(strlen(etcmembers) + strlen(dnencoded) + 2);
  sprintf(memberpath, "%s/%s", etcmembers, dnencoded);

  if (stat(memberpath, &statbuf) != 0)
    {
      fp = fopen(memberpath, "w");
      if (fp == NULL)
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: GridSite\nContent-type: text/html\n");      

          puts("Failed to create record for new member");
          return;
        }
      else 
        {
          fprintf(fp, "<member>\n<addedby>%s</addedby>\n</member>", userdn);
          fclose(fp); 
        }
    }

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite");
  printf("Location: %s%s\n\n", GRST_MEMBERS, MildUrlEncode(&dn[1]));
}

void EditMember(void)
{
  char        *p, *dn, *dnencoded, *memberpath;
  struct tm   *thistm;
  time_t       this_t;
  FILE        *fp;
  struct stat  statbuf;
  xmlDocPtr    thisperson;
  xmlNodePtr   cur;
  xmlChar     *content;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  dn = GetFromCGI("dn");

  if (ReadMemberData(&thisperson, dn))
    {    
      if ((strcmp(userdn, dn) != 0) &&
          (InGroup(GRST_EDITMBRS, userdn) == GRID_ACL_NONE) &&
          ((InGroup(GRST_ADDMBRS, userdn) == GRID_ACL_NONE) ||
           (strcmp(userdn, GetElement(thisperson, "addedby")) != 0)) )
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: GridSite\nContent-type: text/html\n");      

          puts("You do not have permission to add members");
          return;
        }

      PutElement(thisperson, "forenames", GetFromCGI("forenames"));
      PutElement(thisperson, "surname",   GetFromCGI("surname"));

      for (cur = member_elements; cur != NULL; cur = cur->next)
         {
           content = xmlNodeGetContent(cur);
           if (content == NULL) continue;
           if (xmlStrcmp(content,(const xmlChar *) "forenames") == 0) continue;
           if (xmlStrcmp(content,(const xmlChar *) "surname")   == 0) continue;

           PutElement(thisperson, content, GetFromCGI(content));
         }
      
      dnencoded = UrlEncode(dn);  
      memberpath = malloc(strlen(etcmembers) + strlen(dnencoded) + 2);
      sprintf(memberpath, "%s/%s", etcmembers, dnencoded);

      if (xmlSaveFile(memberpath, thisperson) == -1)
        {
          puts("HTTP/1.1 500 Denied");
          printf("Date: %s", asctime(thistm));
          puts("Server: GridSite\nContent-type: text/html\n");      

          printf("Failed to create record for new member %s\n", memberpath);
          return;
        }
    }    

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite");
  printf("Location: %s%s\n\n", GRST_MEMBERS, MildUrlEncode(&dn[1]));
}

void CreateDirectory()
{
  int        n, savederrno;
  char      *p, *dirname, *newpathdir, *oldpathfile, *newpathfile;
  struct tm *thistm;
  time_t     this_t;
  size_t     size_dirname;
  FILE      *fp;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have write permission to that directory");
      return;
    }
    
  dirname = GetFromCGI("dirname");

  p = dirname;
  
  while (*p != '\0') 
       {
         if ((*p == '.') || (*p == '/') || (*p == ' ')) 
           {
             puts("HTTP/1.1 500 Denied");
             printf("Date: %s", asctime(thistm));
             puts("Server: GridSite\nContent-type: text/html\n");      

             puts("<title>Cannot create ");
             for (p = dirname; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
             puts("</title>");

             CatHeadOrFoot(GRST_HEADFILE);

             puts("<center>
              <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
              <tr><td>");  

             puts("<h1>Cannot create  ");
             for (p = dirname; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
             puts("</h1>");

             printf("<p>Will not create directories with names including
                     . / or SPACE\n");
         
             puts("</td></tr></table></center>");

             puts("</td></tr></table></body>");
             return;
           }
           
         ++p;
       }

  newpathdir = malloc(strlen(docroot) + strlen(requestdir) + 
                      strlen(dirname) + 1);

  strcpy(newpathdir, docroot);
  strcat(newpathdir, requestdir);
  strcat(newpathdir, dirname);
    
  if (mkdir(newpathdir, 0755) == 0)
    {
      oldpathfile = malloc(strlen(docroot) + strlen(requestdir) + 
                           strlen(GRST_ACL_FILE) + 2);
      sprintf(oldpathfile, "%s%s/%s", docroot, requestdir, GRST_ACL_FILE);

      newpathfile = malloc(strlen(newpathdir) + strlen(GRST_ACL_FILE) + 2);
      sprintf(newpathfile, "%s/%s", newpathdir, GRST_ACL_FILE);
      
      /* Take ACL from parent directory. At the moment this is different 
         from SlashGrid, where the ACL is initialised as "admin UserDN" */ 
      CopyFile(newpathfile, oldpathfile);
    
      /* everything worked ok so we just redirect to the directory list */
      puts("HTTP/1.1 302 Moved Temporarily");
      printf("Date: %s", asctime(thistm));
      fputs("Server: GridSite\nLocation: ", stdout);

      fputs(requestdir, stdout);
      puts("?list");
      return;      
    }

  savederrno = errno;

  puts("HTTP/1.1 500 Denied");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>Cannot create ");
  for (p = newpathdir; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Cannot create ");
  for (p = newpathdir; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  printf("<p>mkdir() failed with error %d: %s\n", 
            savederrno, strerror(savederrno));
         
  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
}

void StatusPage(char *rawcookiedn, char *cookiedn, char *certdn, 
                xmlDocPtr thisone)
{
  struct tm *thistm;
  time_t     this_t;
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>Website access status</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Website access status</h1>");

  printf("<p>(For help with problems, see the <a href=\"%s\">website help
          pages</a>.)\n", GRST_WEBSITE);

  if (ishttps) puts("<p>You are using the Secure (https) server.");
  else         puts("<p>You are using the Insecure (http) server.");

  puts("You can change servers by following the &quot;Switch to ...&quot; 
        link at the bottom of most pages.");
           
  if (userdn[0] == '\0') 
      puts("<p>You have not passed any valid credentials to this site.");
  else
    {
      printf("<p>Your user identity is <b>%s</b>\n", userdn);

      puts("<p>(Cookies: ");
      if (cookiedn != NULL) 
           printf("<b>%s</b>)\n", cookiedn);
      else puts("none)");
     
      puts("<br>(User Cert: ");
      if (certdn != NULL) 
           printf("<b>%s</b>)\n", certdn);
      else puts("none)");

      printf("<p>Contact details:<br><table>
              <tr><td>Name:</td><td>%s %s</td></tr>
              <tr><td>Email:</td><td>%s</td></tr>
              <tr><td>URL:</td><td>%s</td></tr>
              </table>\n", 
              GetElement(thisone, "forenames"),
              GetElement(thisone, "surname"),
              GetElement(thisone, "email"), 
              GetElement(thisone, "url"));
    }
    
  puts("<p>You can use this form to obtain cookies for the duration of this
           session (ie until you exit your browser.)");

  puts("<table><form action=\"/?setcookie\" method=post>");
  
  if (rawcookiedn != NULL) 
       printf("<tr><td>DN:</td>
               <td><input type=text name=dn size=40 value=\"%s\"></td></tr>", 
                rawcookiedn);
  else puts(  "<tr><td>DN:</td>
               <td><input type=text name=dn size=40 value=\"/O=Grid/O=UKHEP/OU=\"></td></tr>");
  
  puts("<tr><td>Passcode:</td><td><input type=password width=6 name=passcode></td></tr>");  
  puts("<tr><td colspan=2><input type=submit value=\"Get Cookie\"></td></tr>");
  puts("</table></form>");
      
  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
}

void CreationFooter()
{
  int i;

  printf("<form action=\"%s?create\" method=post>\n", requestdir);
  
  puts("<h2>Create a new webpage</h2>");
      
  puts("<table><tr><td>New HTML file:</td>
        <td><input type=text name=filename>
        <input type=hidden name=pagetext 
value=\"&lt;head&gt;
&lt;title&gt;Page title goes here&lt;/title&gt;
&lt;/head&gt;
&lt;body&gt;
&lt;h1&gt;Page heading goes here&lt;/h1&gt;
&lt;p&gt;
A paragraph starts like this and a link to the
&lt;a href=&quot;/&quot;&gt;homepage looks like this&lt;/a&gt;
&lt;/body&gt;\"></td>
        <td><input type=submit value=\"Create\"></td></tr>
        </table></form>");

  printf("<form action=\"%s?createdir\" method=post>\n", requestdir);
  
  puts("<h2>Create a new subdirectory</h2>");
      
  puts("<table><tr><td>New directory:</td>
        <td><input type=text name=dirname></td>
        <td><input type=submit value=\"Create\"></td></tr>
        </table></form>");

  printf("<form action=\"%s?upload\" 
          enctype=\"multipart/form-data\" method=post>\n", requestdir);
  
  puts("<h2>Upload a file / webpage</h2>");

  puts("<table>
        <tr><td>From:</td><td><input type=file name=localfile></td></tr>
        <tr><td>To:</td><td><input type=text name=filename></td></tr>
        <tr><td colspan=2 align=center><input type=submit value=\"Upload\"></td></tr>
        </table></form>");

  puts("<form action=\"./?additem\" method=post>");

  puts("<h2>Add a weblog item</h2>");

  puts("<table>
        <tr><td align=center><input type=submit value=\"Create new item\"></td></tr>
        </table></form>");
}

void DisplayAcl(void)
/* Displays Acl in a table for the current directory */
{
  char *aclfile, *p, oneline[GRST_ROLELEN + GRST_GROUPLEN + 2], *dn,
       *aclgroup, **dirarray;
  int   level, i, ngroups;
  FILE *aclfp;

  aclfile = malloc(strlen(pathtranslated) + strlen(GRST_ACL_FILE) + 1);
  
  strcpy(aclfile, pathtranslated);
  
  p = rindex(aclfile, '/');
  if (p == NULL) return; /* weirdness */  
  p[1] = '\0';
  
  strcat(aclfile, GRST_ACL_FILE);
      
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: Try to open %s in DisplayAcl()", 
               getenv("REMOTE_ADDR"), aclfile);
#endif

  aclfp = fopen(aclfile, "r");
  free(aclfile);
  if (aclfp != NULL)
    {
      puts("<a name=groups><h2>Groups</h2></a>\n<table border=1>");

      while (fgets(oneline, GRST_ROLELEN + GRST_GROUPLEN + 2, aclfp) != NULL)
       {
         p = index(oneline, '\n');
         if (p != NULL) *p = '\0';

         p = index(oneline, ' ');
         if (p != NULL)  
           { 
             *p = '\0';
             aclgroup = &p[1];
           }
         else continue;
         
         if (strcmp(aclgroup, GRID_ACL_ANY_DN) == 0)
              puts("<tr><td align=middle>ANYONE</td>");
         else printf("<tr><td align=middle><a href=\"%s%s\">%s</a></td>\n", 
                              GRST_GROUPS, aclgroup, aclgroup);

         if (aclstatus < GRID_ACL_ADMIN)
             printf("<td>%s</td></tr>\n", oneline);
         else 
           {
             puts("<form action=\"?editacl\" method=post><td align=middle>
                   <small><select name=levelname>");
             
             for (i=0; aclsyms[i] != NULL; ++i)
              if (strcmp(oneline, aclsyms[i]) == 0)
                   printf("<option selected>%s\n", aclsyms[i]);
              else printf("<option>%s\n", aclsyms[i]);
             
             puts("</select>");
             
             printf("<input type=hidden name=group value=\"%s\">\n", aclgroup);
             puts("<input type=submit value=Change></small></td></form></tr>");
           }
       }

      fclose(aclfp);
      puts("</table>");
    }
#ifdef GRST_LOGLOCAL
  else syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: Cannot open a %s in DisplayAcl() %s", 
               getenv("REMOTE_ADDR"), GRST_ACL_FILE, strerror(errno));
#endif
    
  if (aclstatus == GRID_ACL_ADMIN)
    {
      ngroups = 0;      
      dirarray = SortDir(etcgroups);
      if (dirarray != NULL)
        {
          for (i=0; dirarray[i] != NULL; ++i)
           if (InAcl(dirarray[i]) == GRID_ACL_NONE)
             {
               if (ngroups == 0) 
                 printf("<p><form action=\"%s?editacl\" method=post>
                         <select name=group>\n", requestdir);

               printf("<option>%s\n", dirarray[i]);

               ++ngroups;
             }

          free(dirarray);

          if (InAcl(GRID_ACL_ANY_DN) == GRID_ACL_NONE)
            {
              if (ngroups == 0) 
                 printf("<p><form action=\"%s?editacl\" method=post>
                         <select name=group>\n", requestdir);

              printf("<option value=%s>- ANYONE -\n", GRID_ACL_ANY_DN);

              ++ngroups;             
            }

          if (ngroups > 0) 
            puts("</select>
                  <input type=hidden name=levelname value=read>
                  <input type=submit value=\"Add to list\"></form>");
        }
    }

  return;
}

void ListWeblogDirectory(char *dirpath, char *dirarray[], char *querystring)
{
  int        i, last;
  char      *p, *q, *filedirpath, *headerpath;
  struct     stat    statbuf;
  struct tm *thistm;
  time_t     this_t;  
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      
  
  printf("<title>%c", toupper(requesturi[1]));
  if (requesturi[1] != '\0')
      for (p = &requesturi[2]; isalnum(*p); ++p) putchar(*p);
  puts(" Weblog");
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  headerpath = malloc(strlen(dirpath) + 12); 
  strcpy(headerpath, requestdir);
  strcat(headerpath, "/HEADER.txt");
  if (!CatFile(headerpath))
    {
      printf("<h1>%c", toupper(requesturi[1]));
      if (requesturi[1] != '\0')
        for (p = &requesturi[2]; isalnum(*p); ++p) putchar(*p);
      puts(" Weblog");
      puts("</h1>");
    }  

  if (aclstatus >= GRID_ACL_WRITE)
    {
      puts("<p align=center><a href=\"./?additem\"><b>Add new item</b></a>");    
    }

  for (last=0; dirarray[last] != NULL; ++last) ;

  if (dirarray != NULL)
        {
          for (i=last-1; i >= 0; --i)
               {
                 puts("<p>");
                 
                 filedirpath = malloc(strlen(dirpath) + 2 + 
                                          strlen(dirarray[i]));
                 sprintf(filedirpath, "%s/%s", dirpath, dirarray[i]);
                 lstat(filedirpath, &statbuf);
                     
                 p = rindex(dirarray[i], '.');
                 
                 if ((p != NULL) && (strcmp(p, ".wlg") == 0) &&
                                           !S_ISDIR(statbuf.st_mode))
                                           BareWeblogItem(filedirpath, 0);
                 free(filedirpath);
               }
          
          free(dirarray);
        }
}

void ListPlainDirectory(char *dirpath, char *dirarray[])
{
  int        i;
  char      *p, *q, *filedirpath, *mimetype,
             fileftime[81], *headerpath;
  struct     stat    statbuf;
  struct tm *thistm;
  time_t     this_t;  
  
  time(&this_t);
  thistm = gmtime(&this_t); 

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      
  
  puts("<title>List directory");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  headerpath = malloc(strlen(dirpath) + 12); 
  strcpy(headerpath, requestdir);
  strcat(headerpath, "/HEADER.txt");

  puts("<pre>");
  CatFile(headerpath);
  puts("</pre>");

  puts("<h1>Listing of directory");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  puts("<table>");

  puts("<tr>
        <td colspan=3><a href=\"../\">Parent Directory/</a></td>
        <td colspan=4>&nbsp</td>
        </tr>");

  if (dirarray != NULL)
        {
          for (i=0; dirarray[i] != NULL; ++i)
               {
                 filedirpath = malloc(strlen(dirpath) + 2 + 
                                          strlen(dirarray[i]));
                 sprintf(filedirpath, "%s/%s", dirpath, dirarray[i]);
                 lstat(filedirpath, &statbuf);
                 free(filedirpath);
                     
                 strftime(fileftime, sizeof(fileftime), 
                              "%R</td><td align=right>%e %b %y",
                              localtime(&statbuf.st_mtime));
                     
                 if (S_ISDIR(statbuf.st_mode)) 
                      printf("<tr><td>&nbsp</td>
                                  <td>&nbsp</td>
                                  <td>&nbsp</td>
                                  <td><a href=\"%s/\">%s/</a></td>
                                  <td>&nbsp</td>
                                  <td align=right>%s</td></tr>",
                            dirarray[i], dirarray[i], fileftime);
                 else if ((strcmp(dirarray[i], "HEADER.txt") != 0)
                          ||   (aclstatus >= GRID_ACL_WRITE))
                       {
                         mimetype = GetMimeType(dirarray[i]);
                         
                         puts("<tr>");
                             
                         printf("<td><a href=\"%s%s?history\">History</a></td>\n",
                                               requestdir, dirarray[i]);
                             
                         if (((strcmp(mimetype, "text/html") == 0) ||
                                  (strcmp(mimetype, "text/css")  == 0) ||
                                  (strcmp(mimetype, "text/plain") == 0))
                                 && (aclstatus >= GRID_ACL_WRITE))
                            printf("<td><a href=\"%s%s?edit\">Edit</a></td>\n",
                                    requestdir, dirarray[i]);
                         else puts("<td>&nbsp;</td>");       

                         if (aclstatus >= GRID_ACL_WRITE)
                              printf("<td><a 
                               href=\"%s%s?confirmdelete\">Delete</a></td>\n",
                               requestdir, dirarray[i]);
                         else puts("<td>&nbsp;</td>");

                         printf("<td><a href=\"%s\">%s</a></td>
                                     <td align=right>%d</td>
                                     <td align=right>%s</td></tr>",
                                      dirarray[i], dirarray[i], 
                                      statbuf.st_size,
                                      fileftime);
                       }
               }
          
          free(dirarray);
        }

  puts("</table>");
}

void ListDirectory(char *querystring)
{
  int        n ,i, isweblog = 0;
  char      *p, *q, *dirpath, **dirarray;  
  struct tm *thistm;
  time_t     this_t;  
  
  dirpath = strdup(pathtranslated);

  p = rindex(dirpath, '/');
  if (p != NULL) p[0] = '\0';

  dirarray = SortDir(dirpath);
  if ((dirarray != NULL) && (strcmp(querystring, "list") != 0))
    {
      for (i=0; dirarray[i] != NULL; ++i)
         {
           p = rindex(dirarray[i], '.');
           if ((p != NULL) && (strcmp(p, ".wlg") == 0))
             {
               isweblog = 1;
               break;
             }
         }    
    }

  if (isweblog && (aclstatus >= GRID_ACL_READ))
      ListWeblogDirectory(dirpath, dirarray, querystring);
  else if (aclstatus >= GRID_ACL_LIST)          
    {
      ListPlainDirectory(dirpath, dirarray);
      DisplayAcl();
      if (aclstatus >= GRID_ACL_WRITE) CreationFooter(requestdir);
    }
  else
    {
      time(&this_t);
      thistm = gmtime(&this_t); 

      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have list permission to that directory");
      return;
    }
      
  free(dirpath);

  puts("<hr><small>");
  
  if (userdn[0] != '\0') printf("You are %s<br>\n", userdn); 
    
  if (isweblog && (aclstatus >= GRID_ACL_LIST))
       puts("<a href=\"./?list\">List directory</a> .");

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);
  if (!isweblog)
    {
      printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);
      printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a>\n", GRST_GROUPS);
    }
    
  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n",
             GRST_VERSION);

  puts("</small>");

  CatHeadOrFoot(GRST_FOOTFILE);
 
  syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: browse footer sent", getenv("REMOTE_ADDR"));
}

void FileHistory()
{
  int        personret, n, c, filesize, histlen = 0;
  char      *p, *q, *logfilename, *requestdir, *filename, oneline[161], 
             s[81], fileaction[81], *actiondn;
  time_t     this_t, file_t;
  struct tm *thistm, *file_tm;
  FILE      *fp;
  xmlDocPtr  dnperson;
 
  time(&this_t);
  thistm = gmtime(&this_t); 

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>History of file ");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  printf("<h1>History of file <a href=\"");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  fputs("\">", stdout);
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</a></h1>");

  puts("<table>");
 
  puts("<tr><th>Action</th><th>Date</th><th>File size</th><th>Changed
        by</th></tr>");

  requestdir = strdup(requesturi);
  
  filename = rindex(requestdir, '/');
  *filename = '\0';
  ++filename;
  p = index(filename, '?');
  if (p != NULL) *p = '\0';
  if (*filename == '\0') filename = "index.html";

  logfilename = malloc(strlen(docroot) + strlen(requestdir) + 
                                         strlen(filename)   + 8);

  sprintf(logfilename, "%s/%s/.%s.log", docroot, requestdir, filename);

  fp = fopen(logfilename, "r");
  if (fp != NULL)
    {
      while (fgets(oneline, sizeof(oneline), fp) != NULL)
           {
             ++histlen;
             p = rindex(oneline, '\n');
             if (p != NULL) *p = '\0';
             sscanf(oneline, "%80s %d %ld", fileaction, &filesize, &file_t);
             file_tm = localtime(&file_t);     
             strftime(s, sizeof(s), "%R %e %b %y", file_tm); 
             actiondn = index(oneline, '/');

             personret = ReadMemberData(&dnperson, actiondn);
             if (actiondn == NULL) actiondn = "-Not recorded-";
             
             printf("<tr><td>%s</td><td align=right>%s</td>
                         <td align=center>%d</td>\n", 
                         fileaction, s, filesize);

             if (!personret) printf("<td>%s</td>\n", actiondn);
             else            printf("<td><a href=\"%s%s\">%s</td>\n", 
                          GRST_MEMBERS, MildUrlEncode(&actiondn[1]), actiondn);

             puts("</tr>");
           }
     
      fclose(fp);
    }

  free(requestdir);

  if (histlen == 0) 
   puts("<tr><td colspan=4 align=center>No history recorded for this file yet.</td></tr>");

  puts("</table>");

  puts("<hr><small>");
  
  if (userdn[0] != '\0') printf("You are %s<br>\n", userdn); 

  if (aclstatus >= GRID_ACL_WRITE)
      puts("<a href=\"./?list\">List&nbsp;directory</a> .");

  if (ishttps) 
       printf("<a href=\"http://%s%s\">Switch to HTTP</a> .\n", 
             getenv("SERVER_NAME"), requesturi);
  else printf("<a href=\"https://%s%s\">Switch to HTTPS</a> .\n", 
             getenv("SERVER_NAME"), requesturi);

  printf("<a href=\"%s\">Website Help</a> .\n", GRST_WEBSITE);
     
  printf("<a href=\"%s\">List&nbsp;website&nbsp;members</a> .\n", GRST_MEMBERS);
  printf("<a href=\"%s\">List&nbsp;all&nbsp;groups</a> .\n", GRST_GROUPS);

  /* Please dont remove this link to the GridSite page. Thanks, Andrew :) */
  printf("Built with 
           <a href=\"http://www.gridpp.ac.uk/gridsite/\">GridSite</a> %s\n", 
             GRST_VERSION);
     
  puts("</small>");

  CatHeadOrFoot(GRST_FOOTFILE);
}

void EditPage(char *rawpage, char *senderror, char *sendfilename)
{
  char      *p;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      

      puts("You do not have write permission to that page/directory");
      return;
    }

  puts("HTTP/1.1 200 OK");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>Edit");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<h1>Edit");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");
  
  if (senderror != NULL)
      printf("<p><b>Error: %s</b>\n", senderror);

  printf("<form action=\"%s?sendpage\" method=post>\n", requestdir);  

  puts("<p><input type=submit value=\"Save changes\">");

  if (sendfilename != NULL)
      p = sendfilename;
  else
    {
      p = rindex(pathtranslated, '/');
      ++p;
    }
  printf("<input type=text name=filename value=\"%s\">\n", p);

  puts("<p><textarea name=pagetext cols=60 rows=22>");

  for (p=rawpage; *p != '\0'; ++p)
   if      (*p == '<') fputs("&lt;",   stdout);
   else if (*p == '>') fputs("&gt;",   stdout);
   else if (*p == '&') fputs("&amp;",  stdout);
   else if (*p == '"') fputs("&quot;", stdout);
   else putchar(*p);
   
  puts("</textarea>");

  puts("<p><input type=submit value=\"Save changes\">");

  puts("</form>");

  free(rawpage);

  CatHeadOrFoot(GRST_FOOTFILE);
   
  return;
}

void SetCookie()
{
  char      *dn, *passcode;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 

  dn       = GetFromCGI("dn");
  passcode = GetFromCGI("passcode");

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
         "%s: ?setcookie offers dn=%s passcode=%s", getenv("REMOTE_ADDR"), 
         dn, passcode);
#endif

  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  printf("Server: GridSite\n");
  printf("Set-Cookie: gridsiteDN=%s; path=/; expires=Wed, 31-Dec-2003 23:59:59 GMT\n",
          dn);
  printf("Set-Cookie: gridsiteCODE=%s; path=/\n",
          passcode);
  printf("Location: %s\n\n", GRST_STATUS);
}

void SendPage(void)
/* for SendPage, pathtranslated will always be that of the directory, not
   of the file to be written */
{
  int        savederrno = 0, htmltxtflag = 0, n, c;
  char      *p, *pagetext, *filename, *pathfilename, *mimetype,
            *pathfiledatename, *fileext;
  size_t     size_pagetext, size_filename;
  struct tm *thistm;
  struct timeval thisdaytime;
  time_t     this_t;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 
  gettimeofday(&thisdaytime, (struct timezone *) NULL);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("You do not have write permission to that page/directory");
      return;
    }

  filename = GetFromCGI("filename");
  size_filename = strlen(filename) + 1;

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendPage(), filename=%s", getenv("REMOTE_ADDR"), 
              filename);
#endif

  fileext = rindex(filename, '.');
  if (fileext == NULL) fileext = "";

  if (index(filename, '/') != NULL)
    {
      pagetext = GetFromCGI("pagetext");

      EditPage(pagetext, "Error: The file name must not contain /", filename);
      free(filename);
      free(pagetext);
      return;            
    }

  mimetype = GetMimeType(filename);

  if ((strcmp(mimetype, "text/plain")          == 0) ||
      (strcmp(mimetype, "text/html")           == 0) ||
      (strcmp(mimetype, "text/x-weblog-item")  == 0) ||
      (strcmp(mimetype, "text/css")            == 0))   htmltxtflag = 1;
   
  if (!htmltxtflag)    
    {
      pagetext = GetFromCGI("pagetext");

      EditPage(pagetext, 
         "Error: The file extension must be a text/plain, text/html or text/css type",
          filename);
      free(filename);
      free(pagetext);
      return;            
    }

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendPage(), fileext=%s, pathtranslated=%s", 
              getenv("REMOTE_ADDR"), fileext, pathtranslated);
#endif
    
  pathfilename = malloc(strlen(pathtranslated) + 1 + size_filename);
  strcpy(pathfilename, pathtranslated);

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendPage() after strcpy pathfilename",
              getenv("REMOTE_ADDR"));
#endif

  p = rindex(pathfilename, '/');
  p[1] = '\0';

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendPage(), pathfilename1=%s", getenv("REMOTE_ADDR"), 
              pathfilename);
#endif

  pathfiledatename = malloc(strlen(pathfilename) + size_filename + 20);
  sprintf(pathfiledatename, "%s.%s.%ld.%06ld", pathfilename, filename,
           this_t, thisdaytime.tv_usec);

  strcat(pathfilename, filename);
  
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendPage(), pathfilename=%s, pathfiledatename=%s", 
              getenv("REMOTE_ADDR"), pathfilename, pathfiledatename);
#endif

  fp = fopen(pathfiledatename, "w");
  if (fp == NULL) savederrno = errno;
  else
    {
      pagetext = GetFromCGI("pagetext");
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: got pagetext=%s in SendPage()", getenv("REMOTE_ADDR"),
              pagetext);
#endif

      for (p = pagetext; *p != '\0'; ++p) 
                 if (*p != '\r') fputc(*p, fp);

      if (fclose(fp) != 0) savederrno = errno;
      else
        {
#ifdef GRST_LOGLOCAL
          syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: about to unlink in SendPage()", getenv("REMOTE_ADDR"));
#endif

          unlink(pathfilename);

#ifdef GRST_LOGLOCAL
          syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: about to link in SendPage()", getenv("REMOTE_ADDR"));
#endif
          link(pathfiledatename, pathfilename);

          UpdateFileLog("Edit", filename, strlen(pagetext), this_t, 
                                               thisdaytime.tv_usec);
#ifdef GRST_LOGLOCAL
          syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: after UpdateFileLog in SendPage()", getenv("REMOTE_ADDR"));
#endif
        }
        
      free(pagetext);
    }

  free(pathfiledatename);

  if (savederrno == 0)
    {
#ifdef GRST_LOGLOCAL
      syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: finished ok in SendPage()", getenv("REMOTE_ADDR"));
#endif

      /* everything worked ok so we just redirect to the new copy */
      puts("HTTP/1.1 302 Moved Temporarily");
      printf("Date: %s", asctime(thistm));
      fputs("Server: GridSite\nLocation: ", stdout);

      fputs(requestdir, stdout);
 
      if (strcmp(fileext, ".html") == 0) puts(filename);
      else                               puts("?list");
      
      free(filename);
      return;
    }

  puts("HTTP/1.1 500 Internal error");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>Changing file");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Changing file");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  printf("<p>fopen()/fclose() failed with error %d: %s\n", 
              savederrno, strerror(savederrno));

  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
 
  return;
}

void SendItem(void)
/* for SendItem, pathtranslated will always be that of the directory, not
   of the weblog item to be written */
{
  int        savederrno = 0, htmltxtflag = 0, n, c;
  char      *p, *title, *intro, *body, *filename, *pathfilename, *mimetype,
            *pathfiledatename, *fileext;
  size_t     size_pagetext, size_filename;
  struct tm *thistm;
  struct timeval thisdaytime;
  time_t     this_t;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 
  gettimeofday(&thisdaytime, (struct timezone *) NULL);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("You do not have write permission to that directory");
      return;
    }

  filename = GetFromCGI("filename");
  size_filename = strlen(filename) + 1;

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendItem(), filename=%s", getenv("REMOTE_ADDR"), 
              filename);
#endif

  fileext = rindex(filename, '.');
  if (fileext == NULL) fileext = "";

  if (index(filename, '/') != NULL)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("The filename must not contain /");
      return;
    }

  mimetype = GetMimeType(filename);

  if (strcmp(mimetype, "text/x-weblog-item") != 0)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("The file extension must be .wlg for text/x-weblog-item");
      return;
    }

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendItem(), fileext=%s, pathtranslated=%s", 
              getenv("REMOTE_ADDR"), fileext, pathtranslated);
#endif
    
  pathfilename = malloc(strlen(pathtranslated) + 1 + size_filename);
  strcpy(pathfilename, pathtranslated);

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendItem() after strcpy pathfilename",
              getenv("REMOTE_ADDR"));
#endif

  p = rindex(pathfilename, '/');
  p[1] = '\0';

#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendItem(), pathfilename1=%s", getenv("REMOTE_ADDR"), 
              pathfilename);
#endif

  pathfiledatename = malloc(strlen(pathfilename) + size_filename + 20);
  sprintf(pathfiledatename, "%s.%s.%ld.%06ld", pathfilename, filename,
           this_t, thisdaytime.tv_usec);

  strcat(pathfilename, filename);
  
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in SendItem(), pathfilename=%s, pathfiledatename=%s", 
              getenv("REMOTE_ADDR"), pathfilename, pathfiledatename);
#endif

  fp = fopen(pathfiledatename, "w");
  if (fp == NULL) savederrno = errno;
  else
    {
      title = GetFromCGI("title");
      intro = GetFromCGI("intro");
      body  = GetFromCGI("body");

      fprintf(fp, "<weblog-item>\n<author>%s</author>\n<title>%s</title>
              <intro>%s</intro>\n<body>\n", userdn, title, intro);
      
      for (p = body; *p != '\0'; ++p)
         {
           if      (*p == '&') fputs("&amp;" , fp);
           else if (*p == '<') fputs("&lt;"  , fp);
           else if (*p == '>') fputs("&gt;"  , fp);
           else if (*p == '"') fputs("&quot;", fp);
           else                fputc(*p, fp);  
         }
         
      fputs("</body>\n</weblog-item>\n", fp);
      
      if (fclose(fp) != 0) savederrno = errno;
      else
        {
          unlink(pathfilename);
          link(pathfiledatename, pathfilename);
          UpdateFileLog("Edit", filename, strlen(intro) + strlen(body), 
                                            this_t, thisdaytime.tv_usec);
        }        
        
      free(title);
      free(intro);
      free(body);
    }

  free(pathfiledatename);

  if (savederrno == 0)
    {
      /* everything worked ok so we just redirect to the new copy */
      puts("HTTP/1.1 302 Moved Temporarily");
      printf("Date: %s", asctime(thistm));
      printf("Server: GridSite\nLocation: %s%s\n\n", requestdir, filename);

      free(filename);
      return;
    }

  puts("HTTP/1.1 500 Internal error");
  printf("Date: %s", asctime(thistm));
  puts("Server: GridSite\nContent-type: text/html\n");      

  puts("<title>Changing file");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</title>");

  CatHeadOrFoot(GRST_HEADFILE);

  puts("<center>
        <table border=1 bgcolor=\"#FFFFFF\" cellpadding=5 cellspacing=0>
        <tr><td>");  

  puts("<h1>Changing file");
  for (p = requesturi; (*p != '\0') && (*p != '?'); ++p) putchar(*p);
  puts("</h1>");

  printf("<p>fopen()/fclose() failed with error %d: %s\n", 
              savederrno, strerror(savederrno));

  puts("</td></tr></table></center>");

  puts("</td></tr></table></body>");
 
  return;
}

void CreatePage()
{
  int        savederrno = 0;
  char      *p, *pagetext, *filename, *mimetype;
  size_t     size_pagetext, size_filename;
  struct tm *thistm;
  time_t     this_t;

  time(&this_t);
  thistm = gmtime(&this_t); 

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("You do not have write permission to that page/directory");
      return;
    }

  filename = GetFromCGI("filename");
  pagetext = GetFromCGI("pagetext");

  mimetype = GetMimeType(filename);

  if ((index(filename, '/') == NULL) && 
      (strncmp(mimetype, "text/", 5) == 0))
   EditPage(pagetext, NULL, filename);
  else
   EditPage(pagetext, 
    "The filename must be in a text format/mimetype and not contain /",  
      filename);
  
  free(filename);
}

void CreateItem()
{
  char      *p;
  size_t     size_filename;
  struct tm *thistm;
  time_t     this_t;
  FILE      *fp;
  struct timeval thisdaytime;

  time(&this_t);
  thistm = gmtime(&this_t); 
  gettimeofday(&thisdaytime, (struct timezone *) NULL);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("You do not have write permission to that directory");
      return;
    }

  pathtranslated = malloc(strlen(docroot) + strlen(requestdir) + 27);

  sprintf(pathtranslated, "%s/%s/%ld.%ld.wlg", 
                        docroot, requestdir, this_t, thisdaytime.tv_usec);

  fp = fopen(pathtranslated, "w");
  if (fp != NULL)
    {
      fprintf(fp, "<weblog-item>\n<author>%s</author>\n<title></title>\n", 
                  userdn);
      fputs("<intro></intro>\n<body></body>\n</weblog-item>\n", fp);
      fclose(fp);
    }

  EditWeblogItem();
}

int StrCmpShort(char *long_s, char *short_s)
{
  while (*short_s != '\0')
       {
         if (*long_s == '\0') return -1;         

         if (*long_s > *short_s) return +1;
         if (*long_s < *short_s) return -1;
         
         ++short_s;
         ++long_s;
       }

  return 0;
}

#define MALLINC 1000
#define CONTLEN 200
#define MISTUNKNOWN 1
#define MISTUPLOAD  2
#define MISTFILENM  3
#define MAXFILENAME 1000

void StoreUploadFile(char **filebuffer, int *bufferused, char *boundary)
{
  int bufferlen = 0, c, boundarylen;

  *filebuffer = NULL;
  *bufferused = 0;
  boundarylen = strlen(boundary);

  while ((c = getchar()) != EOF)
       {
         ++(*bufferused);
   
         if (*bufferused > bufferlen)
           {
             bufferlen = bufferlen + MALLINC;
             *filebuffer = realloc(*filebuffer, (size_t) bufferlen);

             if (*filebuffer == NULL) return;
           }

         (*filebuffer)[*bufferused - 1] = c;         
         
         if ( (*bufferused > boundarylen)    &&
              (boundary[boundarylen-1] == c) && 
              (boundary[boundarylen-2] == (*filebuffer)[*bufferused - 2]))
           {
#ifdef GRST_LOGLOCAL
             syslog(GRST_LOGLOCAL + LOG_DEBUG, 
                  "%s: StoreUploadFile buffer %s", 
                  getenv("REMOTE_ADDR"), *filebuffer);
#endif
             if (strncmp(&((*filebuffer)[*bufferused-boundarylen]),
                         boundary, boundarylen) == 0)
               {
                 *bufferused = *bufferused - boundarylen - 4;
                 (*filebuffer)[*bufferused] = '\0';
#ifdef GRST_LOGLOCAL
                 syslog(GRST_LOGLOCAL + LOG_DEBUG, 
                  "%s: StoreUploadFile finds boundary %s and returns buffer %s", 
                  getenv("REMOTE_ADDR"), boundary, *filebuffer);
#endif
                 return;
               }
           }           
       }
       
  (*filebuffer)[*bufferused] = '\0';
       
#ifdef GRST_LOGLOCAL
  syslog(GRST_LOGLOCAL + LOG_DEBUG, 
         "%s: StoreUploadFile fails to find boundary", getenv("REMOTE_ADDR"));
#endif

  return;
}

void UploadFile()
{
  char      *contype, *boundary, *p, oneline[200],
             filename[MAXFILENAME] = "", *pathfilename,
            *pathfiledatename, tmpfilename[MAXFILENAME] = "",
            *filebuffer;
  int        mimestate, bufferused = 0, i, savederrno = 0;
  struct tm *thistm;
  struct timeval thisdaytime;
  time_t     this_t;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 
  gettimeofday(&thisdaytime, (struct timezone *) NULL);

  if (aclstatus < GRID_ACL_WRITE)
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("You do not have write permission to that page/directory");
      return;
    }
 
  contype = strdup(getenv("CONTENT_TYPE"));

  if (StrCmpShort(contype, "multipart/form-data; boundary=") != 0)
    {
      free(contype);
      puts("HTTP/1.1 500 Bad format");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("Wrong mime type");
      return;      
    }  

  boundary = &contype[30];
  mimestate = MISTUNKNOWN;
#ifdef GRST_LOGLOCAL
syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: MISTUNKNOWN", getenv("REMOTE_ADDR"));
#endif

syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: before file read in, boundary %s", 
               getenv("REMOTE_ADDR"), boundary);

  while (fgets(oneline, sizeof(oneline), stdin) != NULL)
     {
#ifdef GRST_LOGLOCAL
       syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in UploadFile: read -->%s<--", getenv("REMOTE_ADDR"), oneline);
#endif
       if (*oneline == 13) // MIME has CR/LF line breaks, CR=13
         {
           if      (mimestate == MISTUPLOAD)
             {
#ifdef GRST_LOGLOCAL
               syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: call StoreUploadFile()", 
                      getenv("REMOTE_ADDR"));
#endif
               StoreUploadFile(&filebuffer, &bufferused, boundary);
               mimestate = MISTUNKNOWN;
             }
           else if (mimestate == MISTFILENM)
             {
               fgets(tmpfilename, sizeof(tmpfilename), stdin);
               if (*tmpfilename != 13)
                 {
                   p = index(tmpfilename, 13);
                   *p = '\0';
                   if (*tmpfilename != '\0') strcpy(filename, tmpfilename);
                 }
               mimestate = MISTUNKNOWN;
             }
         }
       else if (StrCmpShort(oneline,             
              "Content-Disposition: form-data; name=\"localfile\"; filename=\"")
                == 0)
         {
           mimestate = MISTUPLOAD;
           if (*filename == '\0') 
             {
               strncpy(filename, &oneline[60], MAXFILENAME-1);

               p = rindex(&oneline[60], '\\');
               if (p != NULL) { ++p ; strncpy(filename, p, MAXFILENAME-1); }

               p = rindex(&oneline[60], '/');
               if (p != NULL) { ++p ; strncpy(filename, p, MAXFILENAME-1); }
                                  
               filename[MAXFILENAME-1] = '\0';
               p = index(filename, '"');
               if (p != NULL) *p = '\0'; 
             }
         }
       else if (StrCmpShort(oneline,
                 "Content-Disposition: form-data; name=\"filename\"") == 0)
         {
           mimestate = MISTFILENM;           
         }      
     }

  free(contype);
  
syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: after file read in", getenv("REMOTE_ADDR"));

  if (filebuffer == NULL)
    {
      puts("HTTP/1.1 500 Failed to get upload document");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("Failed to get upload document");
      return;
    }      

  if (filename[0] == '\0')
    {
      puts("HTTP/1.1 500 Something wrong with filename");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("Something wrong with filename: either not given or included /");
      return;
    }      

  if (index(filename, '/') != NULL)
    {
      puts("HTTP/1.1 500 Filename must not include /");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("<p>Filename must not include /<p>You need to go to the right
           directory and do the upload from there.");
      return;
    }      

  pathfilename = malloc(strlen(pathtranslated) + 2 + strlen(filename));

  strcpy(pathfilename, pathtranslated);
  p = rindex(pathfilename, '/');
  p[1] = '\0';

  pathfiledatename = malloc(strlen(pathfilename) + strlen(filename) + 20);
  sprintf(pathfiledatename, "%s.%s.%ld.%06ld", pathfilename, filename, 
          this_t, thisdaytime.tv_usec);
               
  strcat(pathfilename, filename);

  fp = fopen(pathfiledatename, "w");
  if (fp == NULL) savederrno = errno;
  else
    {
      for (i=0; i < bufferused; ++i) fputc(filebuffer[i], fp);

      if (fclose(fp) != 0) savederrno = errno;
      else
        {
          unlink(pathfilename);
          link(pathfiledatename, pathfilename);

          UpdateFileLog("Upload", filename, bufferused, 
                            this_t, thisdaytime.tv_usec);
        }
   }
	               
  free(pathfilename);

  if (savederrno != 0)
    {
      puts("HTTP/1.1 500 Failed to write out upload document");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("Failed to write out upload document");
      return;
    }
    
  /* everything worked ok so we just redirect to the new copy */
  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  fputs("Server: GridSite\nLocation: ", stdout);
      
  fputs(requestdir, stdout);
  puts("?list");
}

void UploadProxy()
{
  char      *contype, *boundary, *p, oneline[200], *encodeduserdn,
            *pathfilename,
            *pathfiledatename, tmpfilename[MAXFILENAME] = "",
            *filebuffer;
  int        mimestate, bufferused = 0, i, savederrno = 0;
  struct tm *thistm;
  struct timeval thisdaytime;
  time_t     this_t;
  FILE      *fp;

  time(&this_t);
  thistm = gmtime(&this_t); 
  gettimeofday(&thisdaytime, (struct timezone *) NULL);

  if (*userdn == '\0')
    {
      puts("HTTP/1.1 500 Denied");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("You do not have permission to upload a proxy");
      return;
    }
 
  contype = strdup(getenv("CONTENT_TYPE"));

  if (StrCmpShort(contype, "multipart/form-data; boundary=") != 0)
    {
      free(contype);
      puts("HTTP/1.1 500 Bad format");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("Wrong mime type");
      return;      
    }  

  boundary = &contype[30];
  mimestate = MISTUNKNOWN;
#ifdef GRST_LOGLOCAL
syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: MISTUNKNOWN", getenv("REMOTE_ADDR"));
#endif

syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: before file read in, boundary %s", 
               getenv("REMOTE_ADDR"), boundary);

  while (fgets(oneline, sizeof(oneline), stdin) != NULL)
     {
#ifdef GRST_LOGLOCAL
       syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: in UploadFile: read -->%s<--", getenv("REMOTE_ADDR"), oneline);
#endif
       if (*oneline == 13) // MIME has CR/LF line breaks, CR=13
         {
           if      (mimestate == MISTUPLOAD)
             {
#ifdef GRST_LOGLOCAL
               syslog(GRST_LOGLOCAL + LOG_DEBUG, "%s: call StoreUploadFile()", 
                      getenv("REMOTE_ADDR"));
#endif
               umask(077);
               StoreUploadFile(&filebuffer, &bufferused, boundary);
               mimestate = MISTUNKNOWN;
             }
           else if (mimestate == MISTFILENM)
             {
               fgets(tmpfilename, sizeof(tmpfilename), stdin);
               mimestate = MISTUNKNOWN;
             }
         }
       else if (StrCmpShort(oneline,             
              "Content-Disposition: form-data; name=\"localfile\"; filename=\"")
                == 0)
         {
           mimestate = MISTUPLOAD;
         }
       else if (StrCmpShort(oneline,
                 "Content-Disposition: form-data; name=\"filename\"") == 0)
         {
           mimestate = MISTFILENM;           
         }      
     }

  free(contype);
  
syslog(GRST_LOGLOCAL + LOG_DEBUG, 
              "%s: after file read in", getenv("REMOTE_ADDR"));

  if (filebuffer == NULL)
    {
      puts("HTTP/1.1 500 Failed to get uploaded proxy");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("Failed to get uploaded proxy");
      return;
    }      

  encodeduserdn = UrlEncode(userdn);
  pathfilename = malloc(strlen(etcproxies) + strlen(encodeduserdn) + 1);

  strcpy(pathfilename, etcproxies);
  strcat(pathfilename, encodeduserdn);

  fp = fopen(pathfilename, "w");
  if (fp == NULL) savederrno = errno;
  else
    {
      for (i=0; i < bufferused; ++i) fputc(filebuffer[i], fp);

      if (fclose(fp) != 0) savederrno = errno;
   }
	               
  free(pathfilename);

  if (savederrno != 0)
    {
      puts("HTTP/1.1 500 Failed to write out proxy");
      printf("Date: %s", asctime(thistm));
      puts("Server: GridSite\nContent-type: text/html\n");      
      puts("Failed to write out proxy");
      return;
    }
    
  /* everything worked ok so we just redirect to their member page */
  puts("HTTP/1.1 302 Moved Temporarily");
  printf("Date: %s", asctime(thistm));
  printf("Server: GridSite\nLocation: %s%s\n\n", 
                            GRST_MEMBERS, &(MildUrlEncode(userdn))[1] );      
}

